import unittest

import numpy as np
import torch
from torch import nn

from comet.modules.activations import Swish, build_activation


class TestActivations(unittest.TestCase):
    def test_build_activation(self):
        activation = build_activation("ReLU")
        self.assertIsInstance(activation, nn.ReLU)
        activation = build_activation("Tanh")
        self.assertIsInstance(activation, nn.Tanh)
        activation = build_activation("Swish")
        self.assertIsInstance(activation, Swish)

    def test_build_activation_with_nonexistent_function(self):
        with self.assertRaises(Exception) as context:
            build_activation("BadActivation")
        self.assertEqual(
            str(context.exception), "BadActivation invalid activation function."
        )

    def test_swish_result(self):
        input = torch.tensor([-1.0, 2.0, 3.0])
        activation = build_activation("Swish")
        result = activation(input)
        result = torch.round(result * 10 ** 4) / (10 ** 4)  # round to 4 decimals
        expected = torch.tensor(
            [
                -1 * (1 / (1 + np.exp(1))),
                2.0 * (1 / (1 + np.exp(-2))),
                3.0 * (1 / (1 + np.exp(-3))),
            ]
        )
        expected = (
            torch.round(expected * 10 ** 4) / (10 ** 4)
        ).float()  # round to 4 decimals
        self.assertTrue(torch.equal(expected, result))



# -*- coding: utf-8 -*-
import os
import unittest

import torch

from comet.tokenizers import XLMRTextEncoder
from fairseq.models.roberta import XLMRModel
from torchnlp.download import download_file_maybe_extract


class TestXLMRTextEncoder(unittest.TestCase):
    download_file_maybe_extract(
        "https://dl.fbaipublicfiles.com/fairseq/models/xlmr.base.tar.gz",
        directory=os.environ["HOME"] + "/.cache/torch/unbabel_comet/",
        check_files=["xlmr.base/model.pt"],
    )
    xlmr = XLMRModel.from_pretrained(
        os.environ["HOME"] + "/.cache/torch/unbabel_comet/xlmr.base",
        checkpoint_file="model.pt",
    )
    original_vocab = xlmr.task.source_dictionary.__dict__["indices"]
    tokenizer = XLMRTextEncoder(xlmr.encode, original_vocab)

    def test_unk_property(self):
        self.assertEqual(self.tokenizer.unk_index, self.original_vocab["<unk>"])

    def test_pad_property(self):
        self.assertEqual(self.tokenizer.padding_index, self.original_vocab["<pad>"])

    def test_bos_property(self):
        self.assertEqual(self.tokenizer.bos_index, self.original_vocab["<s>"])

    def test_eos_property(self):
        self.assertEqual(self.tokenizer.eos_index, self.original_vocab["</s>"])

    def test_mask_property(self):
        self.assertEqual(self.tokenizer.mask_index, self.original_vocab["<mask>"])

    def test_vocab_property(self):
        self.assertEqual(self.tokenizer.vocab, self.original_vocab)

    def test_vocab_size_property(self):
        self.assertEqual(self.tokenizer.vocab_size, len(self.original_vocab))

    def test_encode(self):
        sentence = "Hello, my dog is cute"
        expected = self.xlmr.encode(sentence)
        result = self.tokenizer.encode(sentence)
        self.assertTrue(torch.equal(expected, result))
        # Make sure the bos and eos tokens were added.
        self.assertEqual(result[0], self.tokenizer.bos_index)
        self.assertEqual(result[-1], self.tokenizer.eos_index)

    def test_batch_encode(self):
        # Test batch_encode.
        batch = ["Hello, my dog is cute", "hello world!"]
        encoded_batch, lengths = self.tokenizer.batch_encode(batch)

        self.assertTrue(torch.equal(encoded_batch[0], self.tokenizer.encode(batch[0])))
        self.assertTrue(
            torch.equal(encoded_batch[1][: lengths[1]], self.tokenizer.encode(batch[1]))
        )
        self.assertEqual(lengths[0], len(self.xlmr.encode("Hello, my dog is cute")))
        self.assertEqual(lengths[1], len(self.xlmr.encode("hello world!")))

        # Check if last sentence is padded.
        self.assertEqual(encoded_batch[1][-1], self.tokenizer.padding_index)
        self.assertEqual(encoded_batch[1][-2], self.tokenizer.padding_index)



from .tokenizer_base import TextEncoderBase
from .hf_tokenizer import HFTextEncoder
from .fastbpe_tokenizer import FastBPEEncoder
from .xlmr_tokenizer import XLMRTextEncoder

__all__ = [
    "XLMRTextEncoder",
    "HFTextEncoder",
    "FastBPEEncoder",
    "TextEncoderBase",
]



# -*- coding: utf-8 -*-
import unittest

from click.testing import CliRunner
from tests.data import DATA_PATH

from comet.cli import download


class TestDownloadCli(unittest.TestCase):
    def setUp(self):
        self.runner = CliRunner()

    def test_download_inexistence_model(self):
        result = self.runner.invoke(
            download,
            ["-m", "inexistence-model", "--saving_path", DATA_PATH],
            catch_exceptions=False,
        )
        self.assertIn(
            "Error: Invalid value for '--model' / '-m': invalid choice: inexistence-model. (choose from ",
            result.stdout,
        )
        self.assertEqual(result.exit_code, 2)

    def test_download_inexistence_corpus(self):
        result = self.runner.invoke(
            download,
            ["-d", "inexistence-corpus", "--saving_path", DATA_PATH],
            catch_exceptions=False,
        )
        self.assertIn(
            "Error: Invalid value for '--data' / '-d': invalid choice: inexistence-corpus. (choose from ",
            result.stdout,
        )
        self.assertEqual(result.exit_code, 2)

    def test_missing_saving_dir(self):
        result = self.runner.invoke(
            download, ["-d", "apequest"], catch_exceptions=False
        )
        self.assertIn("Error: Missing option '--saving_path'.", result.stdout)
        self.assertEqual(result.exit_code, 2)



# -*- coding: utf-8 -*-
import os
from typing import Dict, List

import fastBPE
from torchnlp.download import download_file_maybe_extract

from .tokenizer_base import TextEncoderBase

L93_CODES_URL = "https://dl.fbaipublicfiles.com/laser/models/93langs.fcodes"
L93_VOCAB_URL = "https://dl.fbaipublicfiles.com/laser/models/93langs.fvocab"

L93_CODES_FILE = "93langs.fcodes"
L93_VOCAB_FILE = "93langs.fvocab"

if "HOME" in os.environ:
    saving_directory = os.environ["HOME"] + "/.cache/torch/unbabel_comet/"
else:
    raise Exception("HOME environment variable is not defined.")


class FastBPEEncoder(TextEncoderBase):
    """
    FastBPE LASER tokenizer.

    :param dictionary: LASER vocabulary
    """

    def __init__(self, dictionary: Dict[str, int]) -> None:
        super().__init__()

        if not os.path.exists(saving_directory):
            os.makedirs(saving_directory)

        download_file_maybe_extract(
            L93_CODES_URL, directory=saving_directory, check_files=[L93_CODES_FILE]
        )

        download_file_maybe_extract(
            L93_VOCAB_URL, directory=saving_directory, check_files=[L93_VOCAB_FILE]
        )

        self.bpe = fastBPE.fastBPE(
            saving_directory + L93_CODES_FILE, saving_directory + L93_VOCAB_FILE
        )
        self.bpe_symbol = "@@ "

        # Properties from the base class
        self.stoi = dictionary
        self.itos = [key for key in dictionary.keys()]
        self._pad_index = dictionary["<pad>"]
        self._eos_index = dictionary["</s>"]
        self._unk_index = dictionary["<unk>"]
        self._mask_index = None

    def tokenize(self, sequence: str) -> List[str]:
        return self.bpe.apply([sequence.lower()])[0].split()



# -*- coding: utf-8 -*-
import json
import os
import shutil
import unittest

from click.testing import CliRunner
from comet.cli import score
from comet.models import download_model
from tests.data import DATA_PATH


class TestScoreCli(unittest.TestCase):
    @classmethod
    def tearDownClass(cls):
        shutil.rmtree(DATA_PATH + "emnlp-base-da-ranker")
        shutil.rmtree(DATA_PATH + "wmt-large-da-estimator-1719")
        os.remove(DATA_PATH + "results.json")

    def setUp(self):
        self.runner = CliRunner()

    def test_score_ranker_cpu(self):
        download_model("emnlp-base-da-ranker", DATA_PATH)
        args = [
            "--model",
            DATA_PATH + "emnlp-base-da-ranker/_ckpt_epoch_0.ckpt",
            "-s",
            DATA_PATH + "src.en",
            "-h",
            DATA_PATH + "mt.de",
            "-r",
            DATA_PATH + "ref.de",
            "--cpu",
        ]
        result = self.runner.invoke(score, args, catch_exceptions=False)
        self.assertEqual(result.exit_code, 0)
        self.assertIn("COMET system score: ", result.stdout)

    def test_score_estimator_cpu(self):
        download_model("wmt-large-da-estimator-1719", DATA_PATH)
        args = [
            "--model",
            DATA_PATH + "wmt-large-da-estimator-1719/_ckpt_epoch_1.ckpt",
            "-s",
            DATA_PATH + "src.en",
            "-h",
            DATA_PATH + "mt.de",
            "-r",
            DATA_PATH + "ref.de",
            "--to_json",
            DATA_PATH + "results.json",
            "--cpu",
        ]
        result = self.runner.invoke(score, args, catch_exceptions=False)
        self.assertEqual(result.exit_code, 0)
        self.assertIn(
            "Predictions saved in: {}".format(DATA_PATH + "results.json"), result.stdout
        )
        self.assertTrue(os.path.exists(DATA_PATH + "results.json"))
        with open(DATA_PATH + "results.json") as json_file:
            samples = json.load(json_file)
            for s in samples:
                self.assertIn("predicted_score", s)
        self.assertIn("COMET system score: ", result.stdout)



# -*- coding: utf-8 -*-
import unittest

import numpy as np

import torch
from comet.metrics import RegressionReport, WMTKendall


class TestMetrics(unittest.TestCase):
    def test_regression_report(self):
        report = RegressionReport()
        a = np.array([0, 0, 0, 1, 1, 1, 1])
        b = np.arange(7)
        expected = {
            "pearson": torch.tensor(0.8660254, dtype=torch.float32),
            "kendall": torch.tensor(0.7559289, dtype=torch.float32),
            "spearman": torch.tensor(0.866025, dtype=torch.float32),
        }
        result = report.compute(a, b)
        self.assertDictEqual(
            {k: round(v.item(), 4) for k, v in result.items()},
            {k: round(v.item(), 4) for k, v in expected.items()},
        )

    def test_wmt_kendall(self):
        metric = WMTKendall()

        pos = torch.tensor([0, 0.5, 1])
        neg = torch.tensor([1, 0.5, 0])

        expected = (1 - 2) / (1 + 2)

        self.assertEqual(metric.compute(pos, neg), expected)



# -*- coding: utf-8 -*-
import unittest
from argparse import Namespace

import torch

from comet.models.encoders import XLMREncoder


class TestXLMREncoder(unittest.TestCase):
    """
    There is not much we can test in our encoders...

    The only thing we are interested is in maintaining a well defined interface
    between all encoders.
    """

    hparams = Namespace(pretrained_model="xlmr.base")
    model_base = XLMREncoder.from_pretrained(hparams)

    def test_layerwise_lr(self):
        expected_lr = [
            0.1,
            0.1,
        ]  # LM Head and Last layer with 0.1 and all others with decay
        expected_lr += [0.1 * (0.95 ** i) for i in range(1, 13)]  # LR for other layers
        expected_lr += [
            0.1 * (0.95 ** 13),
            0.1 * (0.95 ** 13),
            0.1 * (0.95 ** 13),
        ]  # LR for embeddings (pos, word, layer_norm)
        parameter_groups = self.model_base.layerwise_lr(0.1, 0.95)
        param_lrs = [params["lr"] for params in parameter_groups[::-1]]
        self.assertListEqual(expected_lr, param_lrs)

    def test_num_layers(self):
        self.assertEqual(self.model_base.num_layers, 13)

    def test_output_units(self):
        self.assertEqual(self.model_base.output_units, 768)

    def test_max_positions(self):
        self.assertEqual(self.model_base.max_positions, 512)

    def test_prepare_sample(self):
        sample = ["hello world, welcome to COMET!", "This is a batch"]

        model_input = self.model_base.prepare_sample(sample)
        self.assertIn("tokens", model_input)
        self.assertIn("lengths", model_input)

        # Sanity Check: This is already checked when testing the tokenizer.
        expected = self.model_base.model.encode(sample[0])
        self.assertTrue(torch.equal(expected, model_input["tokens"][0]))
        self.assertEqual(len(expected), model_input["lengths"][0])

    def test_forward(self):
        sample = ["hello world!", "This is a batch"]
        model_input = self.model_base.prepare_sample(sample)
        model_out = self.model_base(**model_input)
        self.assertIn("wordemb", model_out)
        self.assertIn("sentemb", model_out)
        self.assertIn("all_layers", model_out)
        self.assertIn("mask", model_out)
        self.assertIn("extra", model_out)
        self.assertEqual(len(model_out["all_layers"]), self.model_base.num_layers)
        self.assertEqual(self.model_base.output_units, model_out["sentemb"].size()[1])
        self.assertEqual(self.model_base.output_units, model_out["wordemb"].size()[2])



# -*- coding: utf-8 -*-
r"""
LASER Encoder Model
===================================
    Pretrained LASER Encoder model from Facebook.
    https://github.com/facebookresearch/LASER

    Check the original papers:
        - https://arxiv.org/abs/1704.04154
        - https://arxiv.org/abs/1812.10464
    
    and the original implementation: https://github.com/facebookresearch/LASER
"""
import os
from argparse import Namespace
from typing import Dict

import torch
import torch.nn as nn

from comet.models.encoders.encoder_base import Encoder
from comet.models.utils import convert_padding_direction, sort_sequences
from comet.tokenizers import FastBPEEncoder
from torchnlp.download import download_file_maybe_extract
from torchnlp.utils import lengths_to_mask

# LASER model trained for 93 different languages.
L93_LASER_MODEL_URL = (
    "https://dl.fbaipublicfiles.com/laser/models/bilstm.93langs.2018-12-26.pt"
)
L93_MODEL_NAME = "bilstm.93langs.2018-12-26.pt"


# LASER model trained with europarl parallel data.
EPARL_LASER_MODEL_URL = (
    "https://dl.fbaipublicfiles.com/laser/models/bilstm.eparl21.2018-11-19.pt"
)
EPARL_MODEL_NAME = "bilstm.eparl21.2018-11-19.pt"

if "HOME" in os.environ:
    saving_directory = os.environ["HOME"] + "/.cache/torch/unbabel_comet/"
else:
    raise Exception("HOME environment variable is not defined.")


class LASEREncoder(Encoder):
    """
    Bidirectional LASER Encoder
        
    :param num_embeddings: Size of the vocabulary (73640 BPE tokens).
    :param padding_idx: Index of the padding token in the vocabulary.
    :param embed_dim: Size of the embeddings.
    :param hidden_size: Number of features of the LSTM hidden layer.
    :param num_layers: Number of LSTM stacked layers.
    :param bidirectinal: Flag to initialize a Bidirectional LSTM.
    :param left_pad: If set to True the inputs can be left padded.
        (internaly they will be converted to right padded inputs)
    :param padding_value: Value of the padding token.
    """

    def __init__(
        self,
        num_embeddings: int = 73640,
        padding_idx: int = 1,
        embed_dim: int = 320,
        hidden_size: int = 512,
        num_layers: int = 5,
        bidirectional: bool = True,
        left_pad: bool = True,
        padding_value: float = 0.0,
    ) -> None:
        super().__init__(None)
        self._output_units = hidden_size * 2 if bidirectional else hidden_size
        self._n_layers = num_layers
        self.bidirectional = bidirectional
        self.hidden_size = hidden_size
        self.num_embeddings = num_embeddings
        self.embed_dim = embed_dim

        self.padding_idx = padding_idx
        self.embed_tokens = nn.Embedding(
            num_embeddings, embed_dim, padding_idx=self.padding_idx
        )

        self.lstm = nn.LSTM(
            input_size=embed_dim,
            hidden_size=hidden_size,
            num_layers=num_layers,
            bidirectional=bidirectional,
        )

        self.left_pad = left_pad
        self.padding_value = padding_value

    @property
    def max_positions(self):
        return 4096  # More than that is not recommended

    @property
    def num_layers(self):
        return 1  # In LASER we can only use the last layer

    def freeze_embeddings(self):
        """ Freezes the embedding layer of the network to save some memory while training. """
        for param in self.embed_tokens.parameters():
            param.requires_grad = False

    @classmethod
    def from_pretrained(cls, hparams: Namespace):
        """Function that loads a pretrained LASER encoder and the respective tokenizer.
        
        :param hparams: Namespace.

        :returns: LASER Encoder model
        """
        if not os.path.exists(saving_directory):
            os.makedirs(saving_directory)

        download_file_maybe_extract(
            L93_LASER_MODEL_URL,
            directory=saving_directory,
            check_files=[L93_MODEL_NAME],
        )
        state_dict = torch.load(saving_directory + L93_MODEL_NAME)
        encoder = LASEREncoder(**state_dict["params"])
        encoder.load_state_dict(state_dict["model"])
        encoder.tokenizer = FastBPEEncoder(state_dict["dictionary"])
        return encoder

    def forward(
        self, tokens: torch.Tensor, lengths: torch.Tensor, **kwargs
    ) -> Dict[str, torch.Tensor]:
        """
        Encodes a batch of sequences.

        :param tokens: Torch tensor with the input sequences [batch_size x seq_len].
        :param lengths: Torch tensor with the lenght of each sequence [seq_len].

        :return: Dictionary with `sentemb` (tensor with dims [batch_size x output_units]), `wordemb` 
            (tensor with dims [batch_size x seq_len x output_units]), `mask` (input mask), 
            `all_layers` (List with word_embeddings from all layers, `extra` (tuple with the LSTM outputs, 
            hidden states and cell states).
        """
        self.lstm.flatten_parameters()  # Is it required? should this be in the __init__?
        tokens, lengths, unsorted_idx = sort_sequences(tokens, lengths)

        if self.left_pad:
            # convert left-padding to right-padding
            tokens = convert_padding_direction(
                tokens,
                self.padding_idx,
                left_to_right=True,
            )

        bsz, seqlen = tokens.size()

        # embed tokens
        x = self.embed_tokens(tokens)

        # B x T x C -> T x B x C
        x = x.transpose(0, 1)

        # pack embedded source tokens into a PackedSequence
        packed_x = nn.utils.rnn.pack_padded_sequence(x, lengths.data.tolist())

        # apply LSTM
        if self.bidirectional:
            state_size = 2 * self._n_layers, bsz, self.hidden_size
        else:
            state_size = self._n_layers, bsz, self.hidden_size

        h0 = x.data.new(*state_size).zero_()
        c0 = x.data.new(*state_size).zero_()
        packed_outs, (final_hiddens, final_cells) = self.lstm(packed_x, (h0, c0))

        # unpack outputs and apply dropout
        x, _ = nn.utils.rnn.pad_packed_sequence(
            packed_outs, padding_value=self.padding_value
        )
        assert list(x.size()) == [seqlen, bsz, self.output_units]
        word_embeddings = x

        if self.bidirectional:

            def combine_bidir(outs):
                return torch.cat(
                    [
                        torch.cat([outs[2 * i], outs[2 * i + 1]], dim=0).view(
                            1, bsz, self.output_units
                        )
                        for i in range(self._n_layers)
                    ],
                    dim=0,
                )

            final_hiddens = combine_bidir(final_hiddens)
            final_cells = combine_bidir(final_cells)

        encoder_padding_mask = tokens.eq(self.padding_idx).t()

        # Set padded outputs to -inf so they are not selected by max-pooling
        padding_mask = tokens.eq(self.padding_idx).t().unsqueeze(-1)
        if padding_mask.any():
            x = x.float().masked_fill_(padding_mask, float("-inf")).type_as(x)

        # Build the sentence embedding by max-pooling over the encoder outputs
        sentemb = x.max(dim=0)[0]

        model_out = self.reorder_output(
            encoder_out={
                "sentemb": sentemb,
                "extra": (word_embeddings, final_hiddens, final_cells),
            },
            new_order=unsorted_idx,
        )
        model_out["mask"] = lengths_to_mask(lengths, device=tokens.device)
        model_out["wordemb"] = model_out["extra"][0].transpose(0, 1)
        model_out["all_layers"] = [model_out["wordemb"]]
        return model_out

    def reorder_output(
        self, encoder_out: Dict[str, torch.Tensor], new_order: torch.LongTensor
    ) -> Dict[str, torch.Tensor]:
        """
        Function that reorders the LASER encoder outputs at the batch level.
        
        :param encoder_out: the output of the forward function.
        :param new_order: the new order inside the batch.
        """
        # reorder encoder_out
        encoder_out["extra"] = tuple(
            eo.index_select(1, new_order) for eo in encoder_out["extra"]
        )
        # reorder sentence embeddings
        encoder_out["sentemb"] = encoder_out["sentemb"].index_select(0, new_order)
        return encoder_out



# -*- coding: utf-8 -*-
r"""
Text Encoder 
==============
    Base class difining a common interface between the different tokenizers used
    in COMET.
"""
from typing import List, Dict, Tuple, Iterator

import torch

from torchnlp.encoders import Encoder
from torchnlp.encoders.text import stack_and_pad_tensors
from torchnlp.encoders.text.text_encoder import TextEncoder


class TextEncoderBase(TextEncoder):
    """
    Base class for the specific tokenizers of each model.
    """

    def __init__(self) -> None:
        self.enforce_reversible = False

    @property
    def unk_index(self) -> int:
        """ Returns the index used for the unknown token. """
        return self._unk_index

    @property
    def bos_index(self) -> int:
        """ Returns the index used for the begin-of-sentence token. """
        return self._bos_index

    @property
    def eos_index(self) -> int:
        """ Returns the index used for the end-of-sentence token. """
        return self._eos_index

    @property
    def padding_index(self) -> int:
        """ Returns the index used for padding. """
        return self._pad_index

    @property
    def mask_index(self) -> int:
        """ Returns the index used for masking. """
        return self._mask_index

    @property
    def vocab(self) -> Dict[str, int]:
        """
        Returns:
            dictionary with tokens -> index
        """
        return self.stoi

    @property
    def vocab_size(self) -> int:
        """
        Returns:
            int: Number of tokens in the dictionary.
        """
        return len(self.itos)

    def tokenize(self, sequence: str) -> List[str]:
        """
        Function that tokenizes a string.
        - To be extended by subclasses.
        """
        raise NotImplementedError

    def encode(self, sequence: str) -> torch.Tensor:
        """Encodes a 'sequence'.
        :param sequence: String 'sequence' to encode.

        Returns:
            - torch.Tensor: Encoding of the 'sequence'.
        """
        sequence = super().encode(sequence)
        tokens = self.tokenize(sequence)
        vector = [self.stoi.get(token, self.unk_index) for token in tokens]
        return torch.tensor(vector)

    def batch_encode(
        self, iterator: Iterator[str], dim: int = 0, **kwargs
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        :param iterator (iterator): Batch of text to encode.
        :param dim (int, optional): Dimension along which to concatenate tensors.
        :param **kwargs: Keyword arguments passed to 'encode'.

        Returns
            torch.Tensor, torch.Tensor: Encoded and padded batch of sequences; Original lengths of
                sequences.
        """
        return stack_and_pad_tensors(
            Encoder.batch_encode(self, iterator, **kwargs),
            padding_index=self.padding_index,
            dim=dim,
        )



# -*- coding: utf-8 -*-
import unittest

import torch

from comet.models.encoders import LASEREncoder


class TestLASEREncoder(unittest.TestCase):
    """
    There is not much we can test in our encoders...

    The only thing we are interested is in maintaining a well defined interface
    between all encoders.
    """

    model = LASEREncoder.from_pretrained(None)
    bpe_tokenizer = model.tokenizer

    def test_num_layers(self):
        self.assertEqual(self.model.num_layers, 1)

    def test_output_units(self):
        self.assertEqual(self.model.output_units, 1024)

    def test_prepare_sample(self):
        sample = ["hello world, welcome to COMET!", "This is a batch"]
        model_input = self.model.prepare_sample(sample)
        self.assertIn("tokens", model_input)
        self.assertIn("lengths", model_input)

        expected = self.bpe_tokenizer.encode(sample[0])
        self.assertTrue(torch.equal(expected, model_input["tokens"][0]))
        self.assertEqual(expected.size()[0], model_input["lengths"][0])

    def test_forward(self):
        sample = ["hello world!", "This is a batch"]
        model_input = self.model.prepare_sample(sample)
        model_out = self.model(**model_input)

        self.assertIn("wordemb", model_out)
        self.assertIn("sentemb", model_out)
        self.assertIn("all_layers", model_out)
        self.assertIn("mask", model_out)
        self.assertIn("extra", model_out)

        self.assertEqual(len(model_out["all_layers"]), self.model.num_layers)
        self.assertEqual(self.model.output_units, model_out["sentemb"].size()[1])
        self.assertEqual(self.model.output_units, model_out["wordemb"].size()[2])



# -*- coding: utf-8 -*-
from .estimator_base import Estimator
from .comet_estimator import CometEstimator
from .quality_estimator import QualityEstimator

__all__ = ["Estimator", "CometEstimator", "QualityEstimator"]



# -*- coding: utf-8 -*-
r"""
BERT Encoder
==============
    Pretrained BERT encoder from Hugging Face.
"""
from argparse import Namespace
from typing import Dict

import torch
from transformers import AutoModel

from comet.models.encoders.encoder_base import Encoder
from comet.tokenizers import HFTextEncoder
from torchnlp.utils import lengths_to_mask


class BERTEncoder(Encoder):
    """BERT encoder.
    :param tokenizer: BERT text encoder.
    :param hparams: ArgumentParser.
    """

    def __init__(
        self,
        tokenizer: HFTextEncoder,
        hparams: Namespace,
    ) -> None:
        super().__init__(tokenizer)
        self.model = AutoModel.from_pretrained(hparams.pretrained_model)
        self.model.encoder.output_hidden_states = True
        self._output_units = self.model.config.hidden_size
        self._n_layers = self.model.config.num_hidden_layers + 1
        self._max_pos = self.model.config.max_position_embeddings

    @classmethod
    def from_pretrained(cls, hparams: Namespace) -> Encoder:
        """Function that loads a pretrained encoder from Hugging Face.
        :param hparams: Namespace.

        :return: Encoder model
        """
        tokenizer = HFTextEncoder(model=hparams.pretrained_model)
        model = BERTEncoder(tokenizer=tokenizer, hparams=hparams)
        return model

    def freeze_embeddings(self) -> None:
        """ Frezees the embedding layer of the network to save some memory while training. """
        for param in self.model.embeddings.parameters():
            param.requires_grad = False

    def layerwise_lr(self, lr: float, decay: float):
        """
        :return: List with grouped model parameters with layer-wise decaying learning rate
        """
        # Embedding Layer
        opt_parameters = [
            {
                "params": self.model.embeddings.parameters(),
                "lr": lr * decay ** (self.num_layers),
            }
        ]
        # Last Layer
        last_layer = [{
            "params": self.model.encoder.layer[-1].parameters(),
            "lr": lr
        }]
        # All layers
        opt_parameters += [
            {
                "params": self.model.encoder.layer[l].parameters(),
                "lr": lr * decay ** l,
            }
            for l in range(self.num_layers-2, 0, -1)
        ]
        return opt_parameters + last_layer

    def forward(
        self, tokens: torch.Tensor, lengths: torch.Tensor
    ) -> Dict[str, torch.Tensor]:
        """
        Encodes a batch of sequences.

        :param tokens: Torch tensor with the input sequences [batch_size x seq_len].
        :param lengths: Torch tensor with the lenght of each sequence [seq_len].

        :return: Dictionary with `sentemb` (tensor with dims [batch_size x output_units]), `wordemb` 
            (tensor with dims [batch_size x seq_len x output_units]), `mask` (input mask), 
            `all_layers` (List with word_embeddings from all layers), `extra` (tuple with the 
            last_hidden_state, the pooler_output representing  the entire sentence and the word 
            embeddings for all BERT layers).
        """
        mask = lengths_to_mask(lengths, device=tokens.device)
        last_hidden_states, pooler_output, all_layers = self.model(tokens, mask)
        return {
            "sentemb": pooler_output,
            "wordemb": last_hidden_states,
            "all_layers": all_layers,
            "mask": mask,
            "extra": (last_hidden_states, pooler_output, all_layers),
        }



# -*- coding: utf-8 -*-
r"""
Activation Functions 
==============
    Provides an easy to use interface to initialize different activation functions.
"""
import torch
from torch import nn


def build_activation(activation: str) -> nn.Module:
    """Builder function that returns a nn.module activation function.

    :param activation: string defining the name of the activation function.

    Activations available:
        Swish + every native pytorch activation function.
    """
    if hasattr(nn, activation):
        return getattr(nn, activation)()
    elif activation == "Swish":
        return Swish()
    else:
        raise Exception("{} invalid activation function.".format(activation))


def swish(input: torch.Tensor) -> torch.Tensor:
    """
    Applies Swish element-wise: A self-gated activation function
        swish(x) = x * sigmoid(x)
    """
    return input * torch.sigmoid(input)


class Swish(nn.Module):
    """
    Applies the Swish function element-wise:

        Swish(x) = x * sigmoid(x)

    Shape:
        - Input: (N, *) where * means, any number of additional
          dimensions
        - Output: (N, *), same shape as the input

    References:
        -  Related paper:
        https://arxiv.org/pdf/1710.05941v1.pdf
    """

    def __init__(self) -> None:
        """
        Init method.
        """
        super().__init__()  # init the base class

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        """
        Forward pass of the function.
        """
        return swish(input)



# -*- coding: utf-8 -*-
r"""
Corpora
==============
    Available corpora to train/test COMET models.
"""
import os

import click

from torchnlp.download import download_file_maybe_extract

corpus2download = {
    "apequest": "https://unbabel-experimental-data-sets.s3-eu-west-1.amazonaws.com/comet/hter/apequest.zip",
    "qt21": "https://unbabel-experimental-data-sets.s3-eu-west-1.amazonaws.com/comet/hter/qt21.zip",
    "wmt-metrics": "https://unbabel-experimental-data-sets.s3-eu-west-1.amazonaws.com/comet/da/wmt-metrics.zip",
    "doc-wmt19": "https://unbabel-experimental-data-sets.s3-eu-west-1.amazonaws.com/comet/da/doc-wmt19.zip",
}


def download_corpus(corpus: str, saving_directory: str = None) -> None:
    """Function that downloads a corpus from AWS.
    
    :param corpus: Name of the corpus to be loaded.
    :param saving_directory: RELATIVE path to the saving folder.
    """
    corpus = corpus.lower()
    if not saving_directory:
        saving_directory = "data/"

    if not os.path.exists(saving_directory):
        os.makedirs(saving_directory)

    if os.path.isdir(saving_directory + corpus):
        click.secho(f"{corpus} is already in cache.", fg="yellow")
        return

    elif corpus in corpus2download:
        download_file_maybe_extract(
            corpus2download[corpus],
            directory=saving_directory,
        )

    else:
        raise Exception(f"{corpus} is not a valid corpus!")

    click.secho("Download succeeded.", fg="yellow")
    if os.path.exists(saving_directory + corpus + ".zip"):
        os.remove(saving_directory + corpus + ".zip")

    elif os.path.exists(saving_directory + corpus + ".tar.gz"):
        os.remove(saving_directory + corpus + ".tar.gz")

    else:
        click.secho("Fail to delete compressed file.", fg="red")



# -*- coding: utf-8 -*-
r"""
Encoder Model base
====================
    Module defining the common interface between all pretrained encoder models.
"""
import warnings
from argparse import Namespace
from typing import Dict, List

import torch
import torch.nn as nn

from comet.tokenizers import TextEncoderBase


class Encoder(nn.Module):
    """Base class for an encoder model.

    :param output_units: Number of output features that will be passed to the Estimator.
    """

    def __init__(
        self,
        tokenizer: TextEncoderBase,
    ) -> None:
        super().__init__()
        self.tokenizer = tokenizer

    @property
    def output_units(self):
        """ Max number of tokens the encoder handles. """
        return self._output_units

    @property
    def max_positions(self):
        """ Max number of tokens the encoder handles. """
        return self._max_pos

    @property
    def num_layers(self):
        """ Number of model layers available. """
        return self._n_layers

    @property
    def lm_head(self):
        """ Language modeling head. """
        raise NotImplementedError

    @classmethod
    def from_pretrained(cls, hparams: Namespace):
        """Function that loads a pretrained encoder and the respective tokenizer.

        :return: Encoder model
        """
        raise NotImplementedError

    def check_lengths(self, tokens: torch.Tensor, lengths: torch.Tensor):
        """ Checks if lengths are not exceeded and warns user if so."""
        if lengths.max() > self.max_positions:
            warnings.warn(
                "Encoder max length exceeded ({} > {}).".format(
                    lengths.max(), self.max_positions
                ),
                category=RuntimeWarning,
            )
            lengths[lengths > self.max_positions] = self.max_positions
            tokens = tokens[:, : self.max_positions]

        return tokens, lengths

    def prepare_sample(self, sample: List[str]) -> (torch.Tensor, torch.Tensor):
        """ Receives a list of strings and applies model specific tokenization and vectorization."""
        tokens, lengths = self.tokenizer.batch_encode(sample)
        tokens, lengths = self.check_lengths(tokens, lengths)
        return {"tokens": tokens, "lengths": lengths}

    def freeze(self) -> None:
        """ Frezees the entire encoder network. """
        for param in self.parameters():
            param.requires_grad = False

    def unfreeze(self) -> None:
        """ Unfrezees the entire encoder network. """
        for param in self.parameters():
            param.requires_grad = True

    def freeze_embeddings(self) -> None:
        """ Frezees the embedding layer of the network to save some memory while training. """
        raise NotImplementedError

    def layerwise_lr(self, lr: float, decay: float):
        """
        :return: List with grouped model parameters with layer-wise decaying learning rate
        """
        raise NotImplementedError

    def forward(
        self, tokens: torch.Tensor, lengths: torch.Tensor
    ) -> Dict[str, torch.Tensor]:
        """
        Encodes a batch of sequences.

        :param tokens: Torch tensor with the input sequences [batch_size x seq_len].
        :param lengths: Torch tensor with the lenght of each sequence [seq_len].

        :return: Dictionary with `sentemb` (tensor with dims [batch_size x output_units]), `wordemb` 
            (tensor with dims [batch_size x seq_len x output_units]), `mask` (input mask), 
            `all_layers` (List with word_embeddings from all layers, `extra` (model specific outputs).
        """
        raise NotImplementedError



import unittest

import torch

from comet.models.utils import apply_to_sample, average_pooling, mask_fill, max_pooling


class TestUtils(unittest.TestCase):
    def test_mask_fill(self):
        tokens = torch.tensor([[2, 3, 2], [1, 2, 0]])
        embeddings = torch.tensor(
            [
                [
                    [3.1416, 3.1416, 3.1416],
                    [4.1416, 4.1416, 4.1416],
                    [3.1416, 3.1416, 3.1416],
                ],
                [
                    [1.1416, 1.1416, 1.1416],
                    [3.1416, 3.1416, 3.1416],
                    [0.0000, 0.0000, 0.0000],
                ],
            ]
        )

        expected = torch.tensor(
            [
                [
                    [3.1416, 3.1416, 3.1416],
                    [4.1416, 4.1416, 4.1416],
                    [3.1416, 3.1416, 3.1416],
                ],
                [
                    [1.1416, 1.1416, 1.1416],
                    [3.1416, 3.1416, 3.1416],
                    [-1.000, -1.000, -1.000],
                ],
            ]
        )
        result = mask_fill(-1, tokens, embeddings, 0)
        self.assertTrue(torch.equal(result, expected))

        expected = torch.tensor(
            [
                [
                    [3.1416, 3.1416, 3.1416],
                    [10.000, 10.000, 10.000],
                    [3.1416, 3.1416, 3.1416],
                ],
                [
                    [1.1416, 1.1416, 1.1416],
                    [3.1416, 3.1416, 3.1416],
                    [-1.000, -1.000, -1.000],
                ],
            ]
        )
        result = mask_fill(10.000, tokens, embeddings, 3)
        self.assertTrue(torch.equal(result, expected))

    def test_average_pooling(self):
        tokens = torch.tensor([[2, 2, 2, 2], [2, 2, 2, 0]])
        embeddings = torch.tensor(
            [
                [
                    [3.1416, 3.1416, 3.1416],
                    [3.1416, 3.1416, 3.1416],
                    [3.1416, 3.1416, 3.1416],
                    [3.1416, 3.1416, 3.1416],
                ],
                [
                    [3.1416, 3.1416, 3.1416],
                    [3.1416, 3.1416, 3.1416],
                    [3.1416, 3.1416, 3.1416],
                    [0.0000, 0.0000, 0.0000],
                ],
            ]
        )
        mask = torch.tensor([[1, 1, 1, 1], [1, 1, 1, 0]])

        expected = torch.tensor([[3.1416, 3.1416, 3.1416], [3.1416, 3.1416, 3.1416]])
        result = average_pooling(tokens, embeddings, mask, 0)
        self.assertTrue(torch.equal(result, expected))

    def test_max_pooling(self):
        tokens = torch.tensor([[2, 2, 2, 2], [2, 2, 2, 0]])
        embeddings = torch.tensor(
            [
                [
                    [3.1416, 4.1416, 5.1416],
                    [3.1416, 3.1416, 3.1416],
                    [3.1416, 3.1416, 3.1416],
                    [3.1416, 3.1416, 3.1416],
                ],
                [
                    [6.1416, 3.1416, -3.1416],
                    [3.1416, 3.1416, -1.1416],
                    [3.1416, 7.1416, -3.1416],
                    [0.0000, 0.0000, 0.0000],
                ],
            ]
        )
        expected = torch.tensor([[3.1416, 4.1416, 5.1416], [6.1416, 7.1416, -1.1416]])
        result = max_pooling(tokens, embeddings, 0)
        self.assertTrue(torch.equal(result, expected))

    def test_apply_to_sample(self):
        """ If this works then move_to_cuda and move_to_cpu should work. """

        def to_float(tensor):
            return tensor.float()

        def to_int(tensor):
            return tensor.int()

        sample = {
            "tensor": torch.tensor([2, 2, 2, 2]),
            "another_tensor": torch.tensor([1, 2, 1, 2]),
        }

        sample = apply_to_sample(to_float, sample)
        self.assertTrue(sample["tensor"].is_floating_point())
        self.assertTrue(sample["another_tensor"].is_floating_point())

        sample = apply_to_sample(to_int, sample)
        self.assertFalse(sample["tensor"].is_floating_point())
        self.assertFalse(sample["another_tensor"].is_floating_point())

        sample = [torch.tensor([2, 2, 2, 2]), torch.tensor([1, 2, 1, 2])]
        sample = apply_to_sample(to_float, sample)
        self.assertTrue(sample[0].is_floating_point())
        self.assertTrue(sample[1].is_floating_point())

        sample = apply_to_sample(to_int, sample)
        self.assertFalse(sample[0].is_floating_point())
        self.assertFalse(sample[1].is_floating_point())

        sample = torch.tensor([2, 2, 2, 2])
        sample = apply_to_sample(to_float, sample)
        self.assertTrue(sample.is_floating_point())
        sample = apply_to_sample(to_int, sample)
        self.assertFalse(sample.is_floating_point())



# -*- coding: utf-8 -*-
r"""
COMET command line interface (CLI)
==============
Composed by 4 main commands:
    train       Used to train a machine translation metric.
    score       Uses COMET to score a list of MT outputs.
    download    Used to download corpora or pretrained metric.
"""
import json
import os

import click
import yaml

from comet.corpora import corpus2download, download_corpus
from comet.models import download_model, load_checkpoint, model2download, str2model
from comet.trainer import TrainerConfig, build_trainer
from pytorch_lightning import seed_everything


@click.group()
def comet():
    pass


@comet.command()
@click.option(
    "--config",
    "-f",
    type=click.Path(exists=True),
    required=True,
    help="Path to the configure YAML file",
)
def train(config):
    yaml_file = yaml.load(open(config).read(), Loader=yaml.FullLoader)

    # Build Trainer
    train_configs = TrainerConfig(yaml_file)
    seed_everything(train_configs.seed)
    trainer = build_trainer(train_configs.namespace())

    # Print Trainer parameters into terminal
    result = "Hyperparameters:\n"
    for k, v in train_configs.namespace().__dict__.items():
        result += "{0:30}| {1}\n".format(k, v)
    click.secho(f"{result}", fg="blue", nl=False)

    # Build Model
    try:
        model_config = str2model[train_configs.model].ModelConfig(yaml_file)
        model = str2model[train_configs.model](model_config.namespace())
    except KeyError:
        raise Exception(f"Invalid model {train_configs.model}!")

    result = ""
    for k, v in model_config.namespace().__dict__.items():
        result += "{0:30}| {1}\n".format(k, v)
    click.secho(f"{result}", fg="cyan")

    # Train model
    click.secho(f"{model.__class__.__name__} train starting:", fg="yellow")
    trainer.fit(model)


@comet.command()
@click.option(
    "--model",
    default="wmt-large-da-estimator-1719",
    help="Name of the pretrained model OR path to a model checkpoint.",
    show_default=True,
    type=str,
)
@click.option(
    "--source",
    "-s",
    required=True,
    help="Source segments.",
    type=click.File(),
)
@click.option(
    "--hypothesis",
    "-h",
    required=True,
    help="MT outputs.",
    type=click.File(),
)
@click.option(
    "--reference",
    "-r",
    required=True,
    help="Reference segments.",
    type=click.File(),
)
@click.option(
    "--cuda/--cpu",
    default=True,
    help="Flag that either runs inference on cuda or in cpu.",
    show_default=True,
)
@click.option(
    "--to_json",
    default=False,
    help="Creates and exports model predictions to a JSON file.",
    type=str,
    show_default=True,
)
def score(model, source, hypothesis, reference, cuda, to_json):
    source = [s.strip() for s in source.readlines()]
    hypothesis = [s.strip() for s in hypothesis.readlines()]
    reference = [s.strip() for s in reference.readlines()]
    data = {"src": source, "mt": hypothesis, "ref": reference}
    data = [dict(zip(data, t)) for t in zip(*data.values())]

    model = load_checkpoint(model) if os.path.exists(model) else download_model(model)
    data, scores = model.predict(data, cuda, show_progress=True)

    if isinstance(to_json, str):
        with open(to_json, "w") as outfile:
            json.dump(data, outfile, ensure_ascii=False, indent=4)
        click.secho(f"Predictions saved in: {to_json}.", fg="yellow")

    for i in range(len(scores)):
        click.secho(
            "Segment {} score: {}".format(i, scores[i]), fg="yellow"
        )
    click.secho(
        "COMET system score: {}.".format(sum(scores) / len(scores)), fg="yellow"
    )


@comet.command()
@click.option(
    "--data",
    "-d",
    type=click.Choice(corpus2download.keys(), case_sensitive=False),
    multiple=True,
    help="Public corpora to download.",
)
@click.option(
    "--model",
    "-m",
    type=click.Choice(model2download().keys(), case_sensitive=False),
    multiple=True,
    help="Pretrained models to download.",
)
@click.option(
    "--saving_path",
    type=str,
    help="Relative path to save the downloaded files.",
    required=True,
)
def download(data, model, saving_path):
    for d in data:
        download_corpus(d, saving_path)

    for m in model:
        download_model(m, saving_path)



# -*- coding: utf-8 -*-
r"""
Estimator Base Model
=======================
    Abstract base class used to build new estimator models
    inside COMET.
"""
from argparse import Namespace
from typing import Dict, List, Union

import pandas as pd
import torch
import torch.nn as nn
from tqdm import tqdm

from comet.metrics import RegressionReport
from comet.models.model_base import ModelBase
from comet.models.utils import (average_pooling, max_pooling, move_to_cpu,
                                move_to_cuda)


class Estimator(ModelBase):
    """
    Estimator base class that uses an Encoder to encode sequences.

    :param hparams: Namespace containing the hyperparameters.
    """

    class ModelConfig(ModelBase.ModelConfig):
        """
        Estimator ModelConfig:

        --------------------------- Encoder -----------------------------------------
        
        :param encoder_learning_rate: Learning rate used for the encoder model.

        :param layerwise_decay: Decay for the layer wise learning rates. If 1.0 no decay is applied.

        :param layer: Layer that will be used to extract embeddings. If 'mix' embeddings
            from all layers are combined with a layer-wise attention mechanism

        :param scalar_mix_dropout: If layer='mix' we can regularize layer's importance by
            with a given probability setting that weight to - inf before softmax.

        ------------------------- Feed Forward ---------------------------------------

        :param loss: Loss function to be used (options: binary_xent, mse).

        :param hidden_sizes: String with size of the hidden layers in the feedforward.

        :param activations: Activation functions to be used in the feedforward

        :param dropout: Dropout probability to be applied to the feedforward

        :param final_activation: Activation function to be applied after getting the
            final regression score. Set to False if you wish to perform an 'unbounded' regression.
        """

        encoder_learning_rate: float = 1e-06
        layerwise_decay: float = 1.0
        layer: str = "mix"
        scalar_mix_dropout: float = 0.0

        loss: str = "mse"
        hidden_sizes: str = "1024"
        activations: str = "Tanh"
        dropout: float = 0.1
        final_activation: str = "Sigmoid"

    def __init__(self, hparams: Namespace) -> None:
        super().__init__(hparams)

    def _build_model(self) -> ModelBase:
        """
        Initializes the estimator architecture.
        """
        super()._build_model()
        self.metrics = RegressionReport()

    def _build_loss(self):
        """ Initializes the loss function/s. """
        super()._build_loss()
        if self.hparams.loss == "mse":
            self.loss = nn.MSELoss(reduction="sum")
        elif self.hparams.loss == "binary_xent":
            self.loss = nn.BCELoss(reduction="sum")
        else:
            raise Exception("{} is not a valid loss option.".format(self.hparams.loss))

    def read_csv(self, path: str) -> List[dict]:
        """Reads a comma separated value file.

        :param path: path to a csv file.

        :return: List of records as dictionaries
        """
        df = pd.read_csv(path)
        df = df[["src", "mt", "ref", "score"]]
        df["src"] = df["src"].astype(str)
        df["mt"] = df["mt"].astype(str)
        df["ref"] = df["ref"].astype(str)
        df["score"] = df["score"].astype(float)
        return df.to_dict("records")

    def compute_loss(
        self, model_out: Dict[str, torch.Tensor], targets: Dict[str, torch.Tensor]
    ) -> torch.Tensor:
        """
        Computes Loss value according to a loss function.
        
        :param model_out: model specific output. Must contain a key 'score' with
            a tensor [batch_size x 1] with model predictions
        :param targets: Target score values [batch_size]
        """
        return self.loss(model_out["score"].view(-1), targets["score"])

    def compute_metrics(self, outputs: List[Dict[str, torch.Tensor]]) -> dict:
        """
        Private function that computes metrics of interest based on model predictions and
        respective targets.
        """
        predictions = (
            torch.cat([batch["val_prediction"]["score"].view(-1) for batch in outputs])
            .cpu()
            .numpy()
        )
        targets = (
            torch.cat([batch["val_target"]["score"] for batch in outputs]).cpu().numpy()
        )
        return self.metrics.compute(predictions, targets)

    def get_sentence_embedding(
        self, tokens: torch.Tensor, lengths: torch.Tensor
    ) -> torch.Tensor:
        """Auxiliar function that extracts sentence embeddings for
            a single sentence.
        
        :param tokens: sequences [batch_size x seq_len]
        :param lengths: lengths [batch_size]
        
        :return: torch.Tensor [batch_size x hidden_size]
        """
        # When using just one GPU this should not change behavior
        # but when splitting batches across GPU the tokens have padding
        # from the entire original batch
        if self.trainer and self.trainer.use_dp and self.trainer.num_gpus > 1:
            tokens = tokens[:, : lengths.max()]

        encoder_out = self.encoder(tokens, lengths)

        # for LASER we dont care about the word embeddings
        if self.hparams.encoder_model == "LASER":
            pass

        elif self.scalar_mix:
            embeddings = self.scalar_mix(encoder_out["all_layers"], encoder_out["mask"])

        elif self.layer >= 0 and self.layer < self.encoder.num_layers:
            embeddings = encoder_out["all_layers"][self.layer]

        else:
            raise Exception("Invalid model layer {}.".format(self.layer))

        if self.hparams.pool == "default" or self.hparams.encoder_model == "LASER":
            sentemb = encoder_out["sentemb"]

        elif self.hparams.pool == "max":
            sentemb = max_pooling(
                tokens, embeddings, self.encoder.tokenizer.padding_index
            )

        elif self.hparams.pool == "avg":
            sentemb = average_pooling(
                tokens,
                embeddings,
                encoder_out["mask"],
                self.encoder.tokenizer.padding_index,
            )

        elif self.hparams.pool == "cls":
            sentemb = embeddings[:, 0, :]

        elif self.hparams.pool == "cls+avg":
            cls_sentemb = embeddings[:, 0, :]
            avg_sentemb = average_pooling(
                tokens,
                embeddings,
                encoder_out["mask"],
                self.encoder.tokenizer.padding_index,
            )
            sentemb = torch.cat((cls_sentemb, avg_sentemb), dim=1)
        else:
            raise Exception("Invalid pooling technique.")

        return sentemb

    def predict(
        self,
        samples: List[Dict[str, str]],
        cuda: bool = False,
        show_progress: bool = False,
    ) -> (Dict[str, Union[str, float]], List[float]):
        """Function that runs a model prediction,
        
        :param samples: List of dictionaries with 'mt' and 'ref' keys.
        :param cuda: Flag that runs inference using 1 single GPU.
        :param show_progress: Flag to show progress during inference of multiple examples.

        :return: Dictionary with original samples, predicted scores and langid results for SRC and MT 
            + list of predicted scores
        """
        if self.training:
            self.eval()

        if cuda and torch.cuda.is_available():
            self.to("cuda")

        with torch.no_grad():
            batches = [
                samples[i : i + self.hparams.batch_size]
                for i in range(0, len(samples), self.hparams.batch_size)
            ]
            model_inputs = []
            if show_progress:
                pbar = tqdm(
                    total=len(batches),
                    desc="Preparing batches...",
                    dynamic_ncols=True,
                    leave=None,
                )
            for batch in batches:
                batch = self.prepare_sample(batch, inference=True)
                model_inputs.append(batch)
                if show_progress:
                    pbar.update(1)

            if show_progress:
                pbar.close()

            if show_progress:
                pbar = tqdm(
                    total=len(batches),
                    desc="Scoring hypothesis...",
                    dynamic_ncols=True,
                    leave=None,
                )
            scores = []
            for model_input in model_inputs:
                if cuda and torch.cuda.is_available():
                    model_input = move_to_cuda(model_input)
                    model_out = self.forward(**model_input)
                    model_out = move_to_cpu(model_out)
                else:
                    model_out = self.forward(**model_input)

                model_scores = model_out["score"].numpy().tolist()
                for i in range(len(model_scores)):
                    scores.append(model_scores[i][0])

                if show_progress:
                    pbar.update(1)

            if show_progress:
                pbar.close()

        assert len(scores) == len(samples)
        for i in range(len(scores)):
            samples[i]["predicted_score"] = scores[i]
        return samples, scores

    def document_predict(
        self,
        documents: List[Dict[str, List[str]]],
        cuda: bool = False,
        show_progress: bool = False,
    ) -> (Dict[str, Union[str, float]], List[float]):
        """Function that scores entire documents by processing all segments in parallel.

        :param documents: List of dictionaries with 'mt', 'src' and 'ref' keys where each key is
            a list of segments.
        :param cuda: Flag that runs inference using 1 single GPU.
        :param show_progress: Flag to show progress during inference of multiple examples.

        :return: tuple with Dictionary with original samples and predicted document score, micro 
            average scores, macro average scores.
        """
        if self.training:
            self.eval()

        if cuda and torch.cuda.is_available():
            self.to("cuda")

        inputs, lengths = [], []
        for d in documents:
            d = [dict(zip(d, t)) for t in zip(*d.values())]
            # For very long documents we need to create chunks.
            # (64 sentences per chunk)
            if len(d) > 64:
                document_chunks, document_lengths = [], []
                chunks = [d[i : i + 64] for i in range(0, len(d), 64)]
                for chunk in chunks:
                    chunk = self.prepare_sample(chunk, inference=True)
                    document_lengths.append(chunk["mt_lengths"])
                    if cuda and torch.cuda.is_available():
                        document_chunks.append(chunk)
                lengths.append(torch.cat(document_lengths, dim=0))
                inputs.append(document_chunks)
            else:
                d_input = self.prepare_sample(d, inference=True)
                lengths.append(d_input["mt_lengths"])
                if cuda and torch.cuda.is_available():
                    inputs.append(d_input)

        micro_average, average = [], []
        for doc, seg_lengths in tqdm(
            zip(inputs, lengths),
            total=len(inputs),
            desc="Scoring Documents ...",
            dynamic_ncols=True,
            leave=None,
        ):
            if isinstance(doc, list):
                seg_scores = []
                for chunk in doc:
                    model_output = self.forward(**move_to_cuda(chunk))
                    seg_scores.append(move_to_cpu(model_output)["score"].view(1, -1)[0])
                seg_scores = torch.cat(seg_scores, dim=0)
            else:
                model_output = self.forward(**move_to_cuda(doc))
                seg_scores = move_to_cpu(model_output)["score"].view(1, -1)[0]

            # Invert segment-level scores for HTER
            # seg_scores = torch.ones_like(seg_scores) -  seg_scores
            micro = (seg_scores * seg_lengths).sum() / seg_lengths.sum()
            macro = seg_scores.sum() / seg_scores.size()[0]
            micro_average.append(micro.item())
            average.append(macro.item())

        assert len(micro_average) == len(documents)
        for i in range(len(documents)):
            documents[i]["predicted_score"] = micro_average[i]

        return documents, micro_average, average



# -*- coding: utf-8 -*-
import unittest

import torch
from argparse import Namespace
from comet.schedulers import ConstantLR, WarmupConstant, LinearWarmup


def unwrap_schedule(optimizer, scheduler, num_steps=10):
    lrs = []
    for _ in range(num_steps):
        optimizer.step()
        scheduler.step()
        lrs.append(scheduler.get_last_lr()[0])
    return lrs


class TestSchedulers(unittest.TestCase):
    m = torch.nn.Linear(50, 50)
    optimizer = torch.optim.Adam(m.parameters(), lr=0.001)
    num_steps = 10

    def test_constant_scheduler(self):
        scheduler = ConstantLR.from_hparams(self.optimizer, None)
        lrs = unwrap_schedule(self.optimizer, scheduler, self.num_steps)
        expected = [0.001] * self.num_steps
        self.assertListEqual(lrs, expected)

    def test_warmup_constant(self):
        hparams = Namespace(**{"warmup_steps": 4})
        scheduler = WarmupConstant.from_hparams(self.optimizer, hparams)
        lrs = unwrap_schedule(self.optimizer, scheduler, self.num_steps)
        expected = [
            0.00025,
            0.0005,
            0.00075,
            0.001,
            0.001,
            0.001,
            0.001,
            0.001,
            0.001,
            0.001,
        ]
        self.assertListEqual(lrs, expected)

    def test_linear_warmup(self):
        hparams = Namespace(**{"warmup_steps": 2, "num_training_steps": 10})
        scheduler = LinearWarmup.from_hparams(
            self.optimizer, hparams, num_training_steps=10
        )
        lrs = unwrap_schedule(self.optimizer, scheduler, self.num_steps)
        expected = [
            0.0005,
            0.001,
            0.000875,
            0.00075,
            0.000625,
            0.0005,
            0.000375,
            0.00025,
            0.000125,
            0.0,
        ]
        self.assertListEqual(lrs, expected)



# -*- coding: utf-8 -*-
import os
import pickle
import sys

import click
import pandas as pd
import yaml

import wget
from torchnlp.download import download_file_maybe_extract

from .estimators import CometEstimator, QualityEstimator
from .model_base import ModelBase
from .ranking import CometRanker

str2model = {
    "CometEstimator": CometEstimator,
    "CometRanker": CometRanker,
    # Model that use source only:
    "QualityEstimator": QualityEstimator,
}

MODELS_URL = "https://unbabel-experimental-models.s3.amazonaws.com/comet/share/public-models.yaml"


def get_cache_folder():
    if "HOME" in os.environ:
        cache_directory = os.environ["HOME"] + "/.cache/torch/unbabel_comet/"
        if not os.path.exists(cache_directory):
            os.makedirs(cache_directory)
        return cache_directory
    else:
        raise Exception("HOME environment variable is not defined.")


def model2download(
    saving_directory: str = get_cache_folder(),
    url: str = MODELS_URL,
) -> dict:
    """Download a dictionary with the mapping between models and downloading urls.
    :param saving_directory: RELATIVE path to the saving folder (must end with /).
    Return:
        - dictionary with the mapping between models and downloading urls.
    """
    if not os.path.exists(saving_directory):
        raise FileNotFoundError("The folder to save the model does not exist.")

    if os.path.exists(saving_directory + "available_models.yaml"):
        os.remove(saving_directory + "available_models.yaml")

    file_path = download_file_maybe_extract(
        url=url,
        directory=saving_directory,
        extension="yaml",
    )
    with open(file_path) as fp:
        return yaml.load(fp.read(), Loader=yaml.FullLoader)


def download_model(model: str, saving_directory: str = None) -> ModelBase:
    """Function that loads pretrained models from AWS.
    :param model: Name of the model to be loaded.
    :param saving_directory: RELATIVE path to the saving folder (must end with /).

    Return:
        - Pretrained model.
    """
    if saving_directory is None:
        saving_directory = get_cache_folder()

    if not os.path.exists(saving_directory):
        os.makedirs(saving_directory)

    models = model2download(saving_directory)

    if os.path.isdir(saving_directory + model):
        click.secho(f"{model} is already in cache.", fg="yellow")
        if not model.endswith("/"):
            model += "/"

    elif model not in models.keys():
        raise Exception(f"{model} is not a valid COMET model!")

    elif models[model].startswith("https://"):
        download_file_maybe_extract(models[model], directory=saving_directory)

    else:
        raise Exception("Something went wrong while dowloading the model!")

    if os.path.exists(saving_directory + model + ".zip"):
        os.remove(saving_directory + model + ".zip")

    click.secho("Download succeeded. Loading model...", fg="yellow")
    experiment_folder = saving_directory + model
    checkpoints = [
        file for file in os.listdir(experiment_folder) if file.endswith(".ckpt")
    ]
    checkpoint = checkpoints[-1]
    checkpoint_path = experiment_folder + "/" + checkpoint
    return load_checkpoint(checkpoint_path)


def load_checkpoint(checkpoint: str) -> ModelBase:
    """Function that loads a model from a checkpoint file.
    :param checkpoint: Path to the checkpoint file.

    Returns:
        - COMET Model
    """
    if not os.path.exists(checkpoint):
        raise Exception(f"{checkpoint} file not found!")
    
    tags_csv_file = "/".join(checkpoint.split("/")[:-1] + ["meta_tags.csv"])
    hparam_yaml_file = "/".join(checkpoint.split("/")[:-1] + ["hparams.yaml"])

    if os.path.exists(tags_csv_file):
        # Uggly convertion from older Lightning checkpoints 
        tags = pd.read_csv(
            tags_csv_file, header=None, index_col=0, squeeze=True
        ).to_dict()
        hparams = {}
        for k, v in tags.items():
            if isinstance(v, str) and v.replace('.', '', 1).isdigit():
                hparams[k] = float(v) if '.' in v else int(v)
            else:
                hparams[k] = v
        model = str2model[tags["model"]].load_from_checkpoint(
            checkpoint, hparams=hparams
        )
    elif os.path.exists(hparam_yaml_file):
        with open(hparam_yaml_file) as yaml_file:
            hparams = yaml.load(yaml_file.read(), Loader=yaml.FullLoader)
        model = str2model[hparams["model"]].load_from_checkpoint(
            checkpoint, hparams=hparams
        )
    else:
        raise Exception(
            "[meta_tags.csv|hparams.yaml is missing from the checkpoint folder."
        )

    model.eval()
    model.freeze()
    return model



# -*- coding: utf-8 -*-
import unittest
from argparse import Namespace
from io import StringIO

import numpy as np
import torch
from comet.models import CometEstimator
from comet.models.utils import average_pooling, max_pooling


class TestCometEstimator(unittest.TestCase):

    hparams = Namespace(
        **{
            "encoder_model": "LASER",
            "pretrained_model": None,
            "nr_frozen_epochs": 0,
            "loss": "mse",
            # FeedForward Definition
            "pool": "default",
            "hidden_sizes": "1024",
            "activations": "Tanh",
            "dropout": 0.1,
            "final_activation": False,
        }
    )
    estimator = CometEstimator(hparams)

    def test_read_csv(self):
        csv_file = StringIO("src,ref,mt,score\n" "ola mundo,hi world,hey world!,0.8\n")
        expected = [
            {"src": "ola mundo", "ref": "hi world", "mt": "hey world!", "score": 0.8}
        ]

        data = self.estimator.read_csv(csv_file)
        self.assertListEqual(data, expected)

        # Test ignore extra columns
        csv_file = StringIO(
            "src,ref,mt,score,id\n" "ola mundo,hi world,hey world!,0.8,10299\n"
        )
        data = self.estimator.read_csv(csv_file)
        self.assertListEqual(data, expected)

    def test_compute_metrics(self):
        dummy_outputs = [
            {
                "val_prediction": {
                    "score": torch.tensor(np.array([0, 0, 0, 1, 1, 1, 1])),
                },
                "val_target": {
                    "score": torch.tensor(np.arange(7)),
                },
            }
        ]
        expected = {
            "pearson": torch.tensor(0.8660254, dtype=torch.float32),
            "kendall": torch.tensor(0.7559289, dtype=torch.float32),
            "spearman": torch.tensor(0.866025, dtype=torch.float32),
        }
        metrics = self.estimator.compute_metrics(dummy_outputs)
        self.assertDictEqual(
            {k: round(v.item(), 4) for k, v in metrics.items()},
            {k: round(v.item(), 4) for k, v in expected.items()},
        )

    def test_prepare_sample_to_forward(self):
        """ Test compatability between prepare_sample and forward functions. """
        sample = [
            {"src": "ola mundo", "ref": "hi world", "mt": "hey world!", "score": 0.8},
            {"src": "ola mundo", "ref": "hi world", "mt": "hey world!", "score": 0.8},
        ]

        model_input, target = self.estimator.prepare_sample(sample)
        model_output = self.estimator(**model_input)
        self.assertTrue(model_output["score"].shape[0] == 2)
        self.assertTrue(model_output["score"].shape[1] == 1)

    def test_get_sentence_embedding(self):
        self.estimator.scalar_mix = None
        self.estimator.layer = 0

        # tokens from ["hello world", "how are your?"]
        tokens = torch.tensor([[29733, 4139, 1, 1], [2231, 137, 57374, 8]])
        lengths = torch.tensor([2, 4])
        encoder_out = self.estimator.encoder(tokens, lengths)

        # Expected sentence output with pool = 'max'
        hparams = Namespace(
            # LASER always used default... we need to pretend our encoder is another one
            **{"encoder_model": "other", "pool": "max"}
        )
        self.estimator.hparams = hparams
        # Max pooling is tested in test_utils.py
        expected = max_pooling(tokens, encoder_out["wordemb"], 1)
        sentemb = self.estimator.get_sentence_embedding(tokens, lengths)
        self.assertTrue(torch.equal(sentemb, expected))

        # Expected sentence output with pool = 'avg'
        hparams = Namespace(
            # LASER always used default... we need to pretend our encoder is another one
            **{"encoder_model": "other", "pool": "avg"}
        )
        self.estimator.hparams = hparams
        # AVG pooling is tested in test_utils.py
        expected = average_pooling(
            tokens, encoder_out["wordemb"], encoder_out["mask"], 1
        )
        sentemb = self.estimator.get_sentence_embedding(tokens, lengths)
        self.assertTrue(torch.equal(sentemb, expected))

        # Expected sentence output with pool = 'cls'
        hparams = Namespace(
            # LASER always used default... we need to pretend our encoder is another one
            **{"encoder_model": "other", "pool": "cls"}
        )
        self.estimator.hparams = hparams
        expected = encoder_out["wordemb"][:, 0, :]
        sentemb = self.estimator.get_sentence_embedding(tokens, lengths)
        self.assertTrue(torch.equal(sentemb, expected))

        # Expected sentence output with pool = 'cls+avg'
        hparams = Namespace(
            # LASER always used default... we need to pretend our encoder is another one
            **{"encoder_model": "other", "pool": "cls+avg"}
        )
        self.estimator.hparams = hparams
        cls_embedding = encoder_out["wordemb"][:, 0, :]
        avg_embedding = average_pooling(
            tokens, encoder_out["wordemb"], encoder_out["mask"], 1
        )
        expected = torch.cat((cls_embedding, avg_embedding), dim=1)
        sentemb = self.estimator.get_sentence_embedding(tokens, lengths)
        self.assertTrue(torch.equal(sentemb, expected))

        # Expected sentence output with pool = 'default'
        hparams = Namespace(**{"encoder_model": "LASER", "pool": "default"})
        self.estimator.hparams = hparams
        expected = encoder_out["sentemb"]
        sentemb = self.estimator.get_sentence_embedding(tokens, lengths)
        self.assertTrue(torch.equal(sentemb, expected))



# -*- coding: utf-8 -*-
r"""
Ranking Metrics
==============
    Metrics to evaluate ranking quality of ranker models.
"""
import torch


class WMTKendall():
    def __init__(self):
        self.name = "kendall"

    def compute(
        self, distance_pos: torch.Tensor, distance_neg: torch.Tensor
    ) -> torch.Tensor:
        """Computes the level of concordance, discordance and the WMT kendall tau metric

        :param distance_pos: distance between the positive samples and the anchor/s 
        :param distance_neg: distance between the negative samples and the anchor/s

        :return: Level of agreement, nº of positive sample closer to the anchor
        """
        concordance = torch.sum((distance_pos < distance_neg).float())
        discordance = torch.sum((distance_pos >= distance_neg).float())
        kendall = (concordance - discordance) / (concordance + discordance)
        return kendall



# -*- coding: utf-8 -*-
import unittest
from argparse import Namespace
from io import StringIO

from comet.models import QualityEstimator


class TestQualityEstimator(unittest.TestCase):

    hparams = Namespace(
        **{
            "encoder_model": "LASER",
            "pretrained_model": None,
            "nr_frozen_epochs": 0,
            "loss": "mse",
            # FeedForward Definition
            "pool": "default",
            "hidden_sizes": "1024",
            "activations": "Tanh",
            "dropout": 0.1,
            "final_activation": False,
        }
    )
    estimator = QualityEstimator(hparams)

    def test_read_csv(self):
        csv_file = StringIO("src,mt,score\n" "ola mundo,hey world!,0.8\n")
        expected = [{"src": "ola mundo", "mt": "hey world!", "score": 0.8}]

        data = self.estimator.read_csv(csv_file)
        self.assertListEqual(data, expected)

        # Test ignore extra columns
        csv_file = StringIO("src,mt,score,id\n" "ola mundo,hey world!,0.8,10299\n")
        data = self.estimator.read_csv(csv_file)
        self.assertListEqual(data, expected)

    def test_prepare_sample_to_forward(self):
        """ Test compatability between prepare_sample and forward functions. """
        sample = [
            {"src": "ola mundo", "mt": "hey world!", "score": 0.8},
            {"src": "ola mundo", "mt": "hey world!", "score": 0.8},
        ]

        model_input, target = self.estimator.prepare_sample(sample)
        model_output = self.estimator(**model_input)
        self.assertTrue(model_output["score"].shape[0] == 2)
        self.assertTrue(model_output["score"].shape[1] == 1)



# -*- coding: utf-8 -*-
import torch

from comet.tokenizers import TextEncoderBase


def average_pooling(
    tokens: torch.Tensor,
    embeddings: torch.Tensor,
    mask: torch.Tensor,
    padding_index: int,
) -> torch.Tensor:
    """Average pooling function.

    :param tokens: Word ids [batch_size x seq_length]
    :param embeddings: Word embeddings [batch_size x seq_length x hidden_size]
    :param mask: Padding mask [batch_size x seq_length]
    :param padding_index: Padding value.
    """
    wordemb = mask_fill(0.0, tokens, embeddings, padding_index)
    sentemb = torch.sum(wordemb, 1)
    sum_mask = mask.unsqueeze(-1).expand(embeddings.size()).float().sum(1)
    return sentemb / sum_mask


def max_pooling(
    tokens: torch.Tensor, embeddings: torch.Tensor, padding_index: int
) -> torch.Tensor:
    """Max pooling function.

    :param tokens: Word ids [batch_size x seq_length]
    :param embeddings: Word embeddings [batch_size x seq_length x hidden_size]
    :param padding_index: Padding value.
    """
    return mask_fill(float("-inf"), tokens, embeddings, padding_index).max(dim=1)[0]


def mask_fill(
    fill_value: float,
    tokens: torch.Tensor,
    embeddings: torch.Tensor,
    padding_index: int,
) -> torch.Tensor:
    """
    Function that masks embeddings representing padded elements.
    
    :param fill_value: the value to fill the embeddings belonging to padded tokens.
    :param tokens: The input sequences [bsz x seq_len].
    :param embeddings: word embeddings [bsz x seq_len x hiddens].
    :param padding_index: Index of the padding token.
    """
    padding_mask = tokens.eq(padding_index).unsqueeze(-1)
    return embeddings.float().masked_fill_(padding_mask, fill_value).type_as(embeddings)


def sort_sequences(inputs: torch.Tensor, input_lengths: torch.Tensor):
    """
    Sort sequences according to lengths of the input sequence (descendingly).
    
    :param inputs (Tensor): input sequences, size [B, T, D]
    :param input_lengths (Tensor): length of each sequence, size [B]
    """
    lengths_sorted, sorted_idx = input_lengths.sort(descending=True)
    _, unsorted_idx = sorted_idx.sort()
    return inputs[sorted_idx], lengths_sorted, unsorted_idx


def apply_to_sample(f, sample):
    if hasattr(sample, "__len__") and len(sample) == 0:
        return {}

    def _apply(x):
        if torch.is_tensor(x):
            return f(x)
        elif isinstance(x, dict):
            return {key: _apply(value) for key, value in x.items()}
        elif isinstance(x, list):
            return [_apply(x) for x in x]
        else:
            return x

    return _apply(sample)


def move_to_cuda(sample):
    """ Moves a sample to cuda. Works with dictionaries, tensors and lists. """

    def _move_to_cuda(tensor):
        return tensor.cuda()

    return apply_to_sample(_move_to_cuda, sample)


def move_to_cpu(sample):
    """ Moves a sample to cuda. Works with dictionaries, tensors and lists. """

    def _move_to_cpu(tensor):
        return tensor.cpu()

    return apply_to_sample(_move_to_cpu, sample)


# --------------- LASER auxiliar functions from facebook research ------------------------------
def buffered_arange(max):
    if not hasattr(buffered_arange, "buf"):
        buffered_arange.buf = torch.LongTensor()
    if max > buffered_arange.buf.numel():
        torch.arange(max, out=buffered_arange.buf)
    return buffered_arange.buf[:max]


def convert_padding_direction(
    src_tokens, padding_idx, right_to_left=False, left_to_right=False
):
    assert right_to_left ^ left_to_right
    pad_mask = src_tokens.eq(padding_idx)
    if not pad_mask.any():
        # no padding, return early
        return src_tokens
    if left_to_right and not pad_mask[:, 0].any():
        # already right padded
        return src_tokens
    if right_to_left and not pad_mask[:, -1].any():
        # already left padded
        return src_tokens
    max_len = src_tokens.size(1)
    range = buffered_arange(max_len).type_as(src_tokens).expand_as(src_tokens)
    num_pads = pad_mask.long().sum(dim=1, keepdim=True)
    if right_to_left:
        index = torch.remainder(range - num_pads, max_len)
    else:
        index = torch.remainder(range + num_pads, max_len)
    return src_tokens.gather(1, index)



# -*- coding: utf-8 -*-
r"""
Feed Forward 
==============
    Simple Feed Forward Neural Network module that can be used for classification or regression
"""
import torch
from torch import nn

from comet.modules.activations import build_activation


class FeedForward(nn.Module):
    """
    Feed Forward Neural Network.

    :param in_dim: Number input features.
    :param out_dim: Number of output features. Default is just a score.
    :param hidden_sizes: list with the size of the hidden layers.
        This parameter can also be a string with the sizes splited by a comma.
    :param activations: Name of the activation function to be used in the hidden layers.
    :param final_activation: Name of the final activation function or False if we dont
        want a final activation.
    :param dropout: dropout to be used in the hidden layers.
    """

    def __init__(
        self,
        in_dim: int,
        out_dim: int = 1,
        hidden_sizes: str = "3072,1536,768",
        activations: str = "Sigmoid",
        final_activation: str = "Sigmoid",
        dropout: float = 0.0,
    ) -> None:
        super().__init__()

        if isinstance(hidden_sizes, str):
            hidden_sizes = [int(x) for x in hidden_sizes.split(",")]

        if isinstance(hidden_sizes, int):
            hidden_sizes = [hidden_sizes]

        activation_func = build_activation(activations)

        self.ff = torch.nn.Sequential()
        self.ff.add_module("linear_1", nn.Linear(in_dim, hidden_sizes[0]))
        self.ff.add_module("activation_1", activation_func)
        self.ff.add_module("dropout_1", nn.Dropout(dropout))

        for i in range(1, len(hidden_sizes)):
            self.ff.add_module(
                "linear_{}".format(i + 1),
                nn.Linear(int(hidden_sizes[i - 1]), hidden_sizes[i]),
            )
            self.ff.add_module("activation_{}".format(i + 1), activation_func)
            self.ff.add_module("dropout_{}".format(i + 1), nn.Dropout(dropout))

        self.ff.add_module(
            "linear_{}".format(len(hidden_sizes) + 1),
            nn.Linear(hidden_sizes[-1], int(out_dim)),
        )
        if final_activation:
            final_activation = build_activation(final_activation)
            self.ff.add_module(
                "activation_{}".format(len(hidden_sizes) + 1), final_activation
            )

    def forward(self, in_features: torch.Tensor) -> torch.Tensor:
        return self.ff(in_features)



# -*- coding: utf-8 -*-
r"""
Schedulers
==============
    Leraning Rate schedulers used to train COMET models.
"""
from argparse import Namespace

from torch.optim import Optimizer
from torch.optim.lr_scheduler import LambdaLR


class ConstantPolicy:
    """Policy for updating the LR of the ConstantLR scheduler.
    With this class LambdaLR objects became picklable.
    """

    def __call__(self, *args, **kwargs):
        return 1


class ConstantLR(LambdaLR):
    """
    Constant learning rate schedule

    Wrapper for the huggingface Constant LR Scheduler.
        https://huggingface.co/transformers/v2.1.1/main_classes/optimizer_schedules.html

    :param optimizer: torch.optim.Optimizer
    :param last_epoch:
    """

    def __init__(self, optimizer: Optimizer, last_epoch: int = -1) -> None:
        super(ConstantLR, self).__init__(optimizer, ConstantPolicy(), last_epoch)

    @classmethod
    def from_hparams(
        cls, optimizer: Optimizer, hparams: Namespace, **kwargs
    ) -> LambdaLR:
        """ Initializes a constant learning rate scheduler. """
        return ConstantLR(optimizer)


class WarmupPolicy:
    """Policy for updating the LR of the WarmupConstant scheduler.
    With this class LambdaLR objects became picklable.
    """

    def __init__(self, warmup_steps):
        self.warmup_steps = warmup_steps

    def __call__(self, current_step):
        if current_step < self.warmup_steps:
            return float(current_step) / float(max(1.0, self.warmup_steps))
        return 1.0


class WarmupConstant(LambdaLR):
    """
    Warmup Linear scheduler.
    1) Linearly increases learning rate from 0 to 1 over warmup_steps
        training steps.
    2) Keeps the learning rate constant afterwards.

    :param optimizer: torch.optim.Optimizer
    :param warmup_steps: Linearly increases learning rate from 0 to 1 over warmup_steps.
    :param last_epoch:
    """

    def __init__(
        self, optimizer: Optimizer, warmup_steps: int, last_epoch: int = -1
    ) -> None:
        super(WarmupConstant, self).__init__(
            optimizer, WarmupPolicy(warmup_steps), last_epoch
        )

    @classmethod
    def from_hparams(
        cls, optimizer: Optimizer, hparams: Namespace, **kwargs
    ) -> LambdaLR:
        """ Initializes a constant learning rate scheduler with warmup period. """
        return WarmupConstant(optimizer, hparams.warmup_steps)


class LinearWarmupPolicy:
    """Policy for updating the LR of the LinearWarmup scheduler.
    With this class LambdaLR objects became picklable.
    """

    def __init__(self, warmup_steps, num_training_steps):
        self.num_training_steps = num_training_steps
        self.warmup_steps = warmup_steps

    def __call__(self, current_step):
        if current_step < self.warmup_steps:
            return float(current_step) / float(max(1, self.warmup_steps))
        return max(
            0.0,
            float(self.num_training_steps - current_step)
            / float(max(1, self.num_training_steps - self.warmup_steps)),
        )


class LinearWarmup(LambdaLR):
    """
    Create a schedule with a learning rate that decreases linearly after
    linearly increasing during a warmup period.

    :param optimizer: torch.optim.Optimizer
    :param warmup_steps: Linearly increases learning rate from 0 to 1*learning_rate over warmup_steps.
    :param num_training_steps: Linearly decreases learning rate from 1*learning_rate to 0. over remaining
        t_total - warmup_steps steps.
    :param last_epoch:
    """

    def __init__(
        self,
        optimizer: Optimizer,
        warmup_steps: int,
        num_training_steps: int,
        last_epoch: int = -1,
    ) -> None:
        super(LinearWarmup, self).__init__(
            optimizer, LinearWarmupPolicy(warmup_steps, num_training_steps), last_epoch
        )

    @classmethod
    def from_hparams(
        cls, optimizer: Optimizer, hparams: Namespace, num_training_steps: int
    ) -> LambdaLR:
        """ Initializes a learning rate scheduler with warmup period and decreasing period. """
        return LinearWarmup(optimizer, hparams.warmup_steps, num_training_steps)


str2scheduler = {
    "linear_warmup": LinearWarmup,
    "constant": ConstantLR,
    "warmup_constant": WarmupConstant,
}



# -*- coding: utf-8 -*-
r"""
Regression Metrics
==============
    Metrics to evaluate regression quality of estimator models.
"""
import warnings

import numpy as np
import torch
from scipy.stats import kendalltau, pearsonr, spearmanr


class RegressionReport():
    def __init__(self):
        super().__init__()
        self.metrics = [Pearson(), Kendall(), Spearman()]

    def compute(self, x: np.array, y: np.array) -> float:
        """Computes Kendall correlation.

        :param x: predicted scores.
        :param x: ground truth scores.

        :return: Kendall Tau correlation value.
        """
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            return {metric.name: metric.compute(x, y) for metric in self.metrics}


class Kendall():
    def __init__(self):
        self.name = "kendall"

    def compute(self, x: np.array, y: np.array) -> float:
        """Computes Kendall correlation.

        :param x: predicted scores.
        :param x: ground truth scores.

        :return: Kendall Tau correlation value.
        """
        return torch.tensor(kendalltau(x, y)[0], dtype=torch.float32)


class Pearson():
    def __init__(self):
        self.name = "pearson"

    def compute(self, x: np.array, y: np.array) -> torch.Tensor:
        """Computes Pearson correlation.

        :param x: predicted scores.
        :param x: ground truth scores.

        :return: Pearson correlation value.
        """
        return torch.tensor(pearsonr(x, y)[0], dtype=torch.float32)


class Spearman():
    def __init__(self):
        self.name = "spearman"

    def compute(self, x: np.array, y: np.array) -> float:
        """Computes Spearman correlation.

        :param x: predicted scores.
        :param x: ground truth scores.

        Return:
            - Spearman correlation value.
        """
        return torch.tensor(spearmanr(x, y)[0], dtype=torch.float32)



# -*- coding: utf-8 -*-
from .laser import LASEREncoder
from .bert import BERTEncoder
from .xlmr import XLMREncoder
from .encoder_base import Encoder

str2encoder = {"LASER": LASEREncoder, "BERT": BERTEncoder, "XLMR": XLMREncoder}

__all__ = ["LASEREncoder", "BERTEncoder", "XLMREncoder"]



# -*- coding: utf-8 -*-
import io
import os
import shutil
import unittest
import unittest.mock

from comet.models import (CometEstimator, download_model, load_checkpoint,
                          model2download)
from tests.data import DATA_PATH


class TestDownload(unittest.TestCase):
    @classmethod
    def tearDownClass(cls):
        shutil.rmtree(DATA_PATH + "wmt-base-da-estimator-1719")
        # os.remove(DATA_PATH + "public-models.yaml")

    def test_load_unvalid_checkpoint(self):
        with self.assertRaises(Exception) as context:
            load_checkpoint("folder/that/does/not/exist/")
        self.assertEqual(
            str(context.exception), "folder/that/does/not/exist/ file not found!"
        )
        
    def test_model2download(self):
        model2link = model2download(DATA_PATH)
        self.assertTrue(os.path.exists(DATA_PATH + "public-models.yaml"))
        # Double call should overwrite file.
        model2link = model2download(DATA_PATH)
        self.assertIsInstance(model2link, dict)
        os.remove(DATA_PATH + "public-models.yaml")

    def test_model2download_bad_saving_dir(self):
        self.assertRaises(
            FileNotFoundError, lambda: model2download("folder/that/does/not/exist/")
        )

    @unittest.mock.patch("sys.stdout", new_callable=io.StringIO)
    def test_download_model(self, mock_stdout):
        model = download_model("wmt-base-da-estimator-1719", DATA_PATH)
        self.assertIsInstance(model, CometEstimator)
        self.assertIn("Download succeeded. Loading model...", mock_stdout.getvalue())
        download_model("wmt-base-da-estimator-1719", DATA_PATH)
        self.assertIn("is already in cache.", mock_stdout.getvalue())

    def test_download_wrong_model(self):
        with self.assertRaises(Exception) as context:
            download_model("WrongModel", DATA_PATH)
        self.assertEqual(
            str(context.exception), "WrongModel is not a valid COMET model!"
        )



# -*- coding: utf-8 -*-
r"""
Translation Ranking Base Model
==============================
    Abstract base class used to build new ranking systems inside COMET.
    This task consists of ranking "good" translations above "worse" ones.
"""
from argparse import Namespace
from typing import List

import pandas as pd
import torch
import torch.nn as nn

from comet.models.model_base import ModelBase
from comet.models.utils import average_pooling, max_pooling
from comet.modules.scalar_mix import ScalarMixWithDropout
from comet.metrics import WMTKendall


class RankingBase(ModelBase):
    """
    Ranking Model base class used to fine-tune pretrained models such as XLM-R
    to produce better sentence embeddings by optmizing Triplet Margin Loss.

    :param hparams: Namespace containing the hyperparameters.
    """

    def __init__(
        self,
        hparams: Namespace,
    ) -> None:
        super().__init__(hparams)

    def read_csv(self, path: str) -> List[dict]:
        """Reads a comma separated value file.

        :param path: path to a csv file.

        :return: List of records as dictionaries
        """
        df = pd.read_csv(path)
        df = df[["src", "ref", "pos", "neg"]]
        df["src"] = df["src"].astype(str)
        df["ref"] = df["ref"].astype(str)
        df["pos"] = df["pos"].astype(str)
        df["neg"] = df["neg"].astype(str)
        return df.to_dict("records")

    def _build_loss(self):
        """ Initializes the loss function/s. """
        self.loss = nn.TripletMarginLoss(margin=1.0, p=2)

    def _build_model(self) -> ModelBase:
        """
        Initializes the ranking model architecture.
        """
        super()._build_model()
        self.metrics = WMTKendall()
        if self.hparams.encoder_model != "LASER":
            self.layer = (
                int(self.hparams.layer)
                if self.hparams.layer != "mix"
                else self.hparams.layer
            )

            self.scalar_mix = (
                ScalarMixWithDropout(
                    mixture_size=self.encoder.num_layers,
                    dropout=self.hparams.scalar_mix_dropout,
                    do_layer_norm=True,
                )
                if self.layer == "mix" and self.hparams.pool != "default"
                else None
            )

    def get_sentence_embedding(
        self, tokens: torch.Tensor, lengths: torch.Tensor
    ) -> torch.Tensor:
        """Auxiliar function that extracts sentence embeddings for
            a single sentence.
        
        :param tokens: sequences [batch_size x seq_len]
        :param lengths: lengths [batch_size]
        
        :return: torch.Tensor [batch_size x hidden_size]
        """
        # When using just one GPU this should not change behavior
        # but when splitting batches across GPU the tokens have padding
        # from the entire original batch
        if self.trainer and self.trainer.use_dp and self.trainer.num_gpus > 1:
            tokens = tokens[:, : lengths.max()]

        encoder_out = self.encoder(tokens, lengths)
        # for LASER we dont care about the word embeddings
        if self.hparams.encoder_model == "LASER":
            pass
        elif self.scalar_mix:
            embeddings = self.scalar_mix(encoder_out["all_layers"], encoder_out["mask"])
        elif self.layer >= 0 and self.layer < self.encoder.num_layers:
            embeddings = encoder_out["all_layers"][self.layer]
        else:
            raise Exception("Invalid model layer {}.".format(self.layer))

        if self.hparams.pool == "default" or self.hparams.encoder_model == "LASER":
            sentemb = encoder_out["sentemb"]

        elif self.hparams.pool == "max":
            sentemb = max_pooling(
                tokens, embeddings, self.encoder.tokenizer.padding_index
            )

        elif self.hparams.pool == "avg":
            sentemb = average_pooling(
                tokens,
                embeddings,
                encoder_out["mask"],
                self.encoder.tokenizer.padding_index,
            )

        elif self.hparams.pool == "cls":
            sentemb = embeddings[:, 0, :]

        else:
            raise Exception("Invalid pooling technique.")

        return sentemb


# -*- coding: utf-8 -*-
r"""
Quality Estimator Model
=======================
    Quality Estimator Model estimates a quality score for the hyphotesis (e.g: the MT text)
    by looking only at source and MT.
"""
from argparse import Namespace
from typing import Dict, List, Tuple, Union

import pandas as pd
import torch

from comet.models.estimators import CometEstimator, Estimator
from comet.modules.feedforward import FeedForward
from comet.modules.scalar_mix import ScalarMixWithDropout
from torchnlp.utils import collate_tensors


class QualityEstimator(CometEstimator):
    """
    Estimator class that uses a pretrained encoder to extract features from
    the sequences and then passes those features to a feed forward estimator.

    :param hparams: Namespace containing the hyperparameters.
    """

    def __init__(
        self,
        hparams: Namespace,
    ) -> None:
        super().__init__(hparams)

    def _build_model(self) -> Estimator:
        """
        Initializes the estimator architecture.
        """
        super()._build_model()
        if self.hparams.encoder_model != "LASER":
            self.layer = (
                int(self.hparams.layer)
                if self.hparams.layer != "mix"
                else self.hparams.layer
            )

            self.scalar_mix = (
                ScalarMixWithDropout(
                    mixture_size=self.encoder.num_layers,
                    dropout=self.hparams.scalar_mix_dropout,
                    do_layer_norm=True,
                )
                if self.layer == "mix" and self.hparams.pool != "default"
                else None
            )

        self.ff = FeedForward(
            in_dim=self.encoder.output_units * 4,
            hidden_sizes=self.hparams.hidden_sizes,
            activations=self.hparams.activations,
            dropout=self.hparams.dropout,
        )

    def read_csv(self, path: str) -> List[dict]:
        """Reads a comma separated value file.

        :param path: path to a csv file.

        :return: List of records as dictionaries
        """
        df = pd.read_csv(path)
        df = df[["mt", "src", "score"]]
        df["mt"] = df["mt"].astype(str)
        df["src"] = df["src"].astype(str)
        df["score"] = df["score"].astype(float)
        return df.to_dict("records")

    def prepare_sample(
        self, sample: List[Dict[str, Union[str, float]]], inference: bool = False
    ) -> Union[
        Tuple[Dict[str, torch.Tensor], Dict[str, torch.Tensor]], Dict[str, torch.Tensor]
    ]:
        """
        Function that prepares a sample to input the model.
        :param sample: list of dictionaries.
        :param inference: If set to true prepares only the model inputs.

        :returns: Tuple with 2 dictionaries (model inputs and targets). 
            If `inference=True` returns only the model inputs.
        """
        sample = collate_tensors(sample)
        mt_inputs = self.encoder.prepare_sample(sample["mt"])
        src_inputs = self.encoder.prepare_sample(sample["src"])

        mt_inputs = {"mt_" + k: v for k, v in mt_inputs.items()}
        src_inputs = {"src_" + k: v for k, v in src_inputs.items()}

        inputs = {**mt_inputs, **src_inputs}

        if inference:
            return inputs

        targets = {"score": torch.tensor(sample["score"], dtype=torch.float)}
        return inputs, targets

    def forward(
        self,
        mt_tokens: torch.tensor,
        src_tokens: torch.tensor,
        mt_lengths: torch.tensor,
        src_lengths: torch.tensor,
        **kwargs
    ) -> Dict[str, torch.Tensor]:
        """
        Function that encodes both Source, MT and returns a quality score.

        :param mt_tokens: MT sequences [batch_size x mt_seq_len]
        :param src_tokens: SRC sequences [batch_size x src_seq_len]
        :param mt_lengths: MT lengths [batch_size]
        :param src_lengths: SRC lengths [batch_size]

        :return: Dictionary with model outputs to be passed to the loss function.
        """
        mt_sentemb = self.get_sentence_embedding(mt_tokens, mt_lengths)
        src_sentemb = self.get_sentence_embedding(src_tokens, src_lengths)

        diff_src = torch.abs(mt_sentemb - src_sentemb)
        prod_src = mt_sentemb * src_sentemb

        embedded_sequences = torch.cat(
            (mt_sentemb, src_sentemb, prod_src, diff_src), dim=1
        )
        return {"score": self.ff(embedded_sequences)}



# -*- coding: utf-8 -*-
import unittest
from argparse import Namespace
from io import StringIO

from comet.models import CometRanker
from comet.models.utils import average_pooling, max_pooling

import torch


class TestCometRanker(unittest.TestCase):

    hparams = Namespace(
        **{"encoder_model": "LASER", "pretrained_model": None, "nr_frozen_epochs": 0}
    )
    ranker = CometRanker(hparams)

    def test_read_csv(self):
        csv_file = StringIO(
            "src,ref,pos,neg\n" "ola mundo,hi world,hey world,hey world!\n"
        )
        expected = [
            {
                "src": "ola mundo",
                "ref": "hi world",
                "pos": "hey world",
                "neg": "hey world!",
            }
        ]

        data = self.ranker.read_csv(csv_file)
        self.assertListEqual(data, expected)

        # Test ignore extra columns
        csv_file = StringIO(
            "src,ref,pos,neg,id\n" "ola mundo,hi world,hey world,hey world!,10293\n"
        )
        data = self.ranker.read_csv(csv_file)
        self.assertListEqual(data, expected)

    def test_compute_metrics(self):
        dummy_outputs = [
            {
                "val_prediction": {
                    "src_sentemb": torch.tensor(
                        [[0.4963, 0.7682, 0.0885], [0.1320, 0.3074, 0.6341]]
                    ),
                    "ref_sentemb": torch.tensor(
                        [[0.4901, 0.8964, 0.4556], [0.6323, 0.3489, 0.4017]]
                    ),
                    "pos_sentemb": torch.tensor(
                        [[0.0223, 0.1689, 0.2939], [0.5185, 0.6977, 0.8000]]
                    ),
                    "neg_sentemb": torch.tensor(
                        [[0.1610, 0.2823, 0.6816], [0.9152, 0.3971, 0.8742]]
                    ),
                }
            }
        ]
        # Distance from positive embedding to source: [0.7912, 0.5738]
        # Distance from positive embedding to reference: [0.8800, 0.5415]
        # Harmonic mean: [0.8332, 0.5572]

        # Distance from negative embedding to source: [0.8369, 0.8240]
        # Distance from positive embedding to reference: [0.7325, 0.5528]
        # Harmonic mean: [0.7812, 0.6617]
        expected = {"kendall": torch.tensor(0.0)}
        metrics = self.ranker.compute_metrics(dummy_outputs)
        self.assertDictEqual(metrics, expected)

    def test_prepare_sample_to_forward(self):
        """ Test compatability between prepare_sample and forward functions. """
        sample = [
            {
                "src": "hello world",
                "ref": "ola mundo",
                "pos": "ola mundo",
                "neg": "oi mundo",
            }
        ]
        model_input, target = self.ranker.prepare_sample(sample)
        model_output = self.ranker(**model_input)
        self.assertTrue(model_output["src_sentemb"].shape[0] == 1)
        self.assertTrue(model_output["ref_sentemb"].shape[0] == 1)
        self.assertTrue(model_output["pos_sentemb"].shape[0] == 1)
        self.assertTrue(model_output["neg_sentemb"].shape[0] == 1)
        self.assertTrue(model_output["src_sentemb"].shape[1] == 1024)
        self.assertTrue(model_output["ref_sentemb"].shape[1] == 1024)
        self.assertTrue(model_output["pos_sentemb"].shape[1] == 1024)
        self.assertTrue(model_output["neg_sentemb"].shape[1] == 1024)

    def test_get_sentence_embedding(self):
        self.ranker.scalar_mix = None
        self.ranker.layer = 0

        # tokens from ["hello world", "how are your?"]
        tokens = torch.tensor([[29733, 4139, 1, 1], [2231, 137, 57374, 8]])
        lengths = torch.tensor([2, 4])

        encoder_out = self.ranker.encoder(tokens, lengths)

        # Expected sentence output with pool = 'default'
        hparams = Namespace(**{"encoder_model": "LASER", "pool": "default"})
        self.ranker.hparams = hparams
        expected = encoder_out["sentemb"]
        sentemb = self.ranker.get_sentence_embedding(tokens, lengths)
        self.assertTrue(torch.equal(sentemb, expected))

        # Expected sentence output with pool = 'max'
        hparams = Namespace(
            # LASER always used default... we need to pretend our encoder is another one
            **{"encoder_model": "other", "pool": "max"}
        )
        self.ranker.hparams = hparams
        # Max pooling is tested in test_utils.py
        expected = max_pooling(tokens, encoder_out["wordemb"], 1)
        sentemb = self.ranker.get_sentence_embedding(tokens, lengths)
        self.assertTrue(torch.equal(sentemb, expected))

        # Expected sentence output with pool = 'avg'
        hparams = Namespace(
            # LASER always used default... we need to pretend our encoder is another one
            **{"encoder_model": "other", "pool": "avg"}
        )
        self.ranker.hparams = hparams
        # AVG pooling is tested in test_utils.py
        expected = average_pooling(
            tokens, encoder_out["wordemb"], encoder_out["mask"], 1
        )
        sentemb = self.ranker.get_sentence_embedding(tokens, lengths)
        self.assertTrue(torch.equal(sentemb, expected))

        # Expected sentence output with pool = 'cls'
        hparams = Namespace(
            # LASER always used default... we need to pretend our encoder is another one
            **{"encoder_model": "other", "pool": "cls"}
        )
        self.ranker.hparams = hparams
        expected = encoder_out["wordemb"][:, 0, :]
        sentemb = self.ranker.get_sentence_embedding(tokens, lengths)
        self.assertTrue(torch.equal(sentemb, expected))



# -*- coding: utf-8 -*-
import torch
from transformers import AutoTokenizer

from torchnlp.encoders.text.text_encoder import TextEncoder

from .tokenizer_base import TextEncoderBase


class HFTextEncoder(TextEncoderBase):
    """
    Wrapper arround transformers AutoTokenizer:
        - https://huggingface.co/transformers/model_doc/auto.html#autotokenizer

    :param model: model to be used.
    """

    def __init__(self, model: str) -> None:
        super().__init__()
        self.tokenizer = AutoTokenizer.from_pretrained(model)
        # Properties from the base class
        self.stoi = self.tokenizer.get_vocab()
        self.itos = {v: k for k, v in self.stoi.items()}
        self._bos_index = self.tokenizer.cls_token_id
        self._pad_index = self.tokenizer.pad_token_id
        self._eos_index = self.tokenizer.sep_token_id
        self._unk_index = self.tokenizer.unk_token_id
        self._mask_index = self.tokenizer.mask_token_id

    def encode(self, sequence: str) -> torch.Tensor:
        """Encodes a 'sequence'.
        :param sequence: String 'sequence' to encode.

        Returns:
            - torch.Tensor: Encoding of the 'sequence'.
        """
        sequence = TextEncoder.encode(self, sequence)
        return torch.tensor(self.tokenizer.encode(sequence))



# -*- coding: utf-8 -*-
import io
import os
import shutil
import unittest

from tests.data import DATA_PATH

from comet.corpora import download_corpus


class TestDownloadCorpus(unittest.TestCase):
    @unittest.mock.patch("sys.stdout", new_callable=io.StringIO)
    def test_download_apequest(self, mock_stdout):
        download_corpus("apequest", DATA_PATH)
        self.assertTrue(os.path.isdir(DATA_PATH + "apequest"))
        self.assertIn("Download succeeded.", mock_stdout.getvalue())
        download_corpus("apequest", DATA_PATH)
        self.assertIn("apequest is already in cache.", mock_stdout.getvalue())
        shutil.rmtree(DATA_PATH + "apequest")

    @unittest.mock.patch("sys.stdout", new_callable=io.StringIO)
    def test_download_qt21(self, mock_stdout):
        download_corpus("qt21", DATA_PATH)
        self.assertTrue(os.path.isdir(DATA_PATH + "qt21"))
        self.assertIn("Download succeeded.", mock_stdout.getvalue())
        shutil.rmtree(DATA_PATH + "qt21")

    @unittest.mock.patch("sys.stdout", new_callable=io.StringIO)
    def test_download_wmt(self, mock_stdout):
        download_corpus("wmt-metrics", DATA_PATH)
        self.assertTrue(os.path.isdir(DATA_PATH + "wmt-metrics"))
        self.assertIn("Download succeeded.", mock_stdout.getvalue())
        shutil.rmtree(DATA_PATH + "wmt-metrics")

    @unittest.mock.patch("sys.stdout", new_callable=io.StringIO)
    def test_download_docwmt19(self, mock_stdout):
        download_corpus("doc-wmt19", DATA_PATH)
        self.assertTrue(os.path.isdir(DATA_PATH + "doc-wmt19"))
        self.assertIn("Download succeeded.", mock_stdout.getvalue())
        shutil.rmtree(DATA_PATH + "doc-wmt19")



# -*- coding: utf-8 -*-
r"""
Model Base
==============
    Abstract base class used to build new modules inside COMET. 
    This class is just an extention of PyTorch Lightning main module:
    https://pytorch-lightning.readthedocs.io/en/0.8.4/lightning-module.html
"""
from argparse import Namespace
from os import path
from typing import Dict, Generator, List, Tuple, Union

import click
import numpy as np
import pandas as pd
import torch
from torch.utils.data import DataLoader, RandomSampler, Subset

import pytorch_lightning as ptl
from comet.models.encoders import Encoder, str2encoder
from comet.schedulers import str2scheduler


class ModelBase(ptl.LightningModule):
    """
    Extends PyTorch Lightning with a common structure and interface
    that will be shared across all architectures.

    :param hparams: Namespace with hyper-parameters
    """

    class ModelConfig:
        """
        The ModelConfig class is used to define model hyper-parameters that
        are used to initialize our Lightning Modules. These parameters are
        then overwritted with the values defined in the YAML file and coverted
        to a Namespace to initialize the model.

        :param model: Model class name (to be replaced with the model specified in the YAML)

        -------------------- Training Parameters -------------------------

        :param batch_size: Batch size used during training.

        :param nr_frozen_epochs: Number of epochs we keep the encoder model frozen.

        :param keep_embeddings_frozen: Keeping the embeddings frozen is a usefull way to save some GPU memory usage.
            This is critical to fine-tune large models in GPUs with less than 32GB memory.

        -------------------- Optimizer Parameters -------------------------

        :param optimizer: Optimizer class to be used.

        :param learning_rate: Overall learning rate.

        -------------------- Scheduler Parameters -------------------------

        :param scheduler: Scheduler class to be used.

        :param warmup_steps: Warmup steps (only used for schedulers with warmup period).

        -------------------- Architecture Parameters -------------------------

        :param encoder_model: Encoder class to be used.

        :param pretrained_model: Encoder checkpoint (e.g. xlmr.base vs xlmr.large)

        :param pool: Pooling technique to extract the sentence embeddings.
            Options: {max, avg, default, cls} where default uses the `default` sentence embedding
            returned by the encoder (e.g. BERT pooler_output) and `cls` is the first token of the
            sequence and depends on the selected layer.

        :param load_weights: Loads weights from a checkpoint file that match the architecture.

        -------------------- Data Parameters -------------------------

        :param train_path: Path to the training data.

        :param val_path: Path to the validation data.

        :param test_path: Path to the test data.

        :param loader_workers: Number of workers used to load and tokenize data during training.

        :param monitor: Metric to be displayed in tqdm bar. Same as trainer monitor flag!
        """

        model: str = None

        # Training details
        batch_size: int = 8
        nr_frozen_epochs: int = 0
        keep_embeddings_frozen: bool = False

        # Optimizer
        optimizer: str = "Adam"
        learning_rate: float = 1e-05

        # Scheduler
        scheduler: str = "constant"
        warmup_steps: int = None

        # Architecture Definition
        encoder_model: str = "XLMR"
        pretrained_model: str = "xlmr.base"
        pool: str = "avg"
        load_weights: str = False

        # Data
        train_path: str = None
        val_path: str = None
        test_path: str = None
        loader_workers: int = 8

        monitor: str = "kendall"

        def __init__(self, initial_data: dict) -> None:
            for key in initial_data:
                if hasattr(self, key):
                    setattr(self, key, initial_data[key])

        def namespace(self) -> Namespace:
            return Namespace(
                **{
                    name: getattr(self, name)
                    for name in dir(self)
                    if not callable(getattr(self, name)) and not name.startswith("__")
                }
            )

    def __init__(self, hparams: Namespace) -> None:
        super(ModelBase, self).__init__()
        if isinstance(hparams, dict):
            self.hparams = Namespace(**hparams)
        else:
            self.hparams = hparams
        self.encoder = self._build_encoder()

        # Model initialization
        self._build_model()

        # Loss criterion initialization.
        self._build_loss()

        # The encoder always starts in a frozen state.
        if self.hparams.nr_frozen_epochs > 0:
            self._frozen = True
            self.freeze_encoder()
        else:
            self._frozen = False

        if (
            hasattr(self.hparams, "keep_embeddings_frozen")
            and self.hparams.keep_embeddings_frozen
        ):
            self.encoder.freeze_embeddings()

        self.nr_frozen_epochs = self.hparams.nr_frozen_epochs

    def _build_loss(self):
        """ Initializes the loss function/s. """
        pass

    def _build_model(self) -> ptl.LightningModule:
        """
        Initializes the estimator architecture.
        """
        # Compatibility with previous COMET versions
        if (
            hasattr(self.hparams, "load_weights")
            and self.hparams.load_weights
            and path.exists(self.hparams.load_weights)
        ):
            click.secho(f"Loading weights from {self.hparams.load_weights}", fg="red")
            self.load_weights(self.hparams.load_weights)

    def _build_encoder(self) -> Encoder:
        """
        Initializes the encoder.
        """
        try:
            return str2encoder[self.hparams.encoder_model].from_pretrained(self.hparams)
        except KeyError:
            raise Exception(f"{self.hparams.encoder_model} invalid encoder model!")

    def _build_optimizer(self, parameters: Generator) -> torch.optim.Optimizer:
        """
        Initializes the Optimizer.

        :param parameters: Module.parameters.
        """
        if hasattr(torch.optim, self.hparams.optimizer):
            return getattr(torch.optim, self.hparams.optimizer)(
                params=parameters, lr=self.hparams.learning_rate
            )
        else:
            raise Exception(f"{self.hparams.optimizer} invalid optimizer!")

    def _build_scheduler(
        self, optimizer: torch.optim.Optimizer
    ) -> torch.optim.lr_scheduler.LambdaLR:
        """
        Initializes the Scheduler.

        :param optimizer: PyTorch optimizer
        """
        self.epoch_total_steps = len(self.train_dataset) // (
            self.hparams.batch_size * max(1, self.trainer.num_gpus)
        )
        self.total_steps = self.epoch_total_steps * float(self.trainer.max_epochs)
        try:
            return {
                "scheduler": str2scheduler[self.hparams.scheduler].from_hparams(
                    optimizer, self.hparams, num_training_steps=self.total_steps
                ),
                "interval": "step",  # called after each training step
            }
        except KeyError:
            raise Exception(f"{self.hparams.scheduler} invalid scheduler!")

    def read_csv(self, path: str) -> List[dict]:
        """Reads a comma separated value file.

        :param path: path to a csv file.

        :return: List of records as dictionaries
        """
        df = pd.read_csv(path)
        return df.to_dict("records")

    def freeze_encoder(self) -> None:
        """ Freezes the encoder layer. """
        self.encoder.freeze()

    def unfreeze_encoder(self) -> None:
        """ un-freezes the encoder layer. """
        if self._frozen:
            if self.trainer.is_global_zero:
                click.secho("\nEncoder model fine-tuning", fg="red")
            self.encoder.unfreeze()
            self._frozen = False

            if (
                hasattr(self.hparams, "keep_embeddings_frozen")
                and self.hparams.keep_embeddings_frozen
            ):
                self.encoder.freeze_embeddings()

    def on_epoch_end(self):
        """ Hook used to unfreeze encoder during training. """
        if self.current_epoch + 1 >= self.nr_frozen_epochs and self._frozen:
            self.unfreeze_encoder()
            self._frozen = False

    def predict(
        self, samples: Dict[str, str]
    ) -> (Dict[str, Union[str, float]], List[float]):
        """Function that runs a model prediction,
        
        :param samples: dictionary with expected model sequences.
            You can also pass a list of dictionaries to predict an entire batch.

        :return: Dictionary with input samples + scores and list with just the scores.
        """
        pass

    def forward(self, *args, **kwargs) -> Dict[str, torch.Tensor]:
        """
        PyTorch Forward.
        
        :return: Dictionary with model outputs to be passed to the loss function.
        """
        pass

    def compute_loss(
        self, model_out: Dict[str, torch.Tensor], targets: Dict[str, torch.Tensor]
    ) -> torch.Tensor:
        """
        Computes Loss value according to a loss function.
        
        :param model_out: model specific output.
        :param targets: Target score values [batch_size]
        """
        pass

    def prepare_sample(
        self, sample: List[Dict[str, Union[str, float]]], inference: bool = False
    ) -> Union[
        Tuple[Dict[str, torch.Tensor], Dict[str, torch.Tensor]], Dict[str, torch.Tensor]
    ]:
        """
        Function that prepares a sample to input the model.
        
        :param sample: List of dictionaries.
        :param inference: If set to true prepares only the model inputs.

        :returns: Tuple with 2 dictionaries (model inputs and targets). If `inference=True`
            returns only the model inputs.
        """
        pass

    def configure_optimizers(
        self,
    ) -> Tuple[List[torch.optim.Optimizer], List[torch.optim.lr_scheduler.LambdaLR]]:
        """
        Function for setting up the optimizers and the schedulers to be used during training.

        :returns: List with as many optimizers as we need and a list with the respective schedulers.
        """
        optimizer = self._build_optimizer(self.parameters())
        scheduler = self._build_scheduler(optimizer)
        return [optimizer], [scheduler]

    def compute_metrics(
        self, outputs: List[Dict[str, torch.Tensor]]
    ) -> Dict[str, torch.Tensor]:
        """
        Function that computes metrics of interest based on the list of outputs
        you defined in validation_step.
        """
        pass

    def training_step(
        self,
        batch: Tuple[Dict[str, torch.Tensor], Dict[str, torch.Tensor]],
        batch_nb: int,
        *args,
        **kwargs,
    ) -> Dict[str, torch.Tensor]:
        """
        Runs one training step.
        This usually consists in the forward function followed by the loss function.

        :param batch: The output of your prepare_sample function.
        :param batch_nb: Integer displaying which batch this is.

        :returns: dictionary containing the loss and the metrics to be added to the lightning logger.
        """
        batch_input, batch_target = batch
        batch_prediction = self.forward(**batch_input)
        loss_value = self.compute_loss(batch_prediction, batch_target)

        # in DP mode (default) make sure if result is scalar, there's another dim in the beginning
        if self.trainer.use_dp or self.trainer.use_ddp2:
            loss_value = loss_value.unsqueeze(0)

        if (
            self.nr_frozen_epochs < 1.0
            and self.nr_frozen_epochs > 0.0
            and batch_nb > self.epoch_total_steps * self.nr_frozen_epochs
        ):
            self.unfreeze_encoder()
            self._frozen = False

        self.log('train_loss', loss_value, on_step=True, on_epoch=True)
        return loss_value

    def validation_step(
        self,
        batch: Tuple[Dict[str, torch.Tensor], Dict[str, torch.Tensor]],
        batch_nb: int,
        dataloader_idx: int,
    ) -> Dict[str, torch.Tensor]:
        """
        Similar to the training step but with the model in eval mode.

        :param batch: The output of your prepare_sample function.
        :param batch_nb: Integer displaying which batch this is.
        :param dataloader_idx: Integer displaying which dataloader this is.

        :returns: dictionary passed to the validation_end function.
        """
        batch_input, batch_target = batch
        batch_prediction = self.forward(**batch_input)
        loss_value = self.compute_loss(batch_prediction, batch_target)

        # in DP mode (default) make sure if result is scalar, there's another dim in the beginning
        if self.trainer.use_dp or self.trainer.use_ddp2:
            loss_value = loss_value.unsqueeze(0)

        return {
            "val_loss": loss_value,
            "val_prediction": batch_prediction,
            "val_target": batch_target,
        }

    def validation_epoch_end(
        self, outputs: List[Dict[str, torch.Tensor]]
    ) -> Dict[str, Dict[str, torch.Tensor]]:
        """
        Function that takes as input a list of dictionaries returned by the validation_step
            and measures the model performance accross the entire validation set.

        :param outputs:
        
        :returns: Dictionary with metrics to be added to the lightning logger.
        """
        train_outs, val_outs = outputs
        train_loss = torch.stack([x["val_loss"] for x in train_outs]).mean()
        val_loss = torch.stack([x["val_loss"] for x in val_outs]).mean()

        # Store Metrics for Reporting...
        val_metrics = self.compute_metrics(val_outs)
        val_metrics["avg_loss"] = val_loss
        self.log_dict(val_metrics, prog_bar=True)

        train_metrics = self.compute_metrics(train_outs)
        train_metrics["avg_loss"] = train_loss
        self.log_dict({"train_" + k: v for k, v in train_metrics.items()})

    def test_step(
        self,
        batch: Tuple[Dict[str, torch.Tensor], Dict[str, torch.Tensor]],
        batch_nb: int,
        *args,
        **kwargs,
    ) -> Dict[str, torch.Tensor]:
        """ Redirects to the validation_step function """
        return self.validation_step(batch, batch_nb, 0)

    def test_epoch_end(
        self, outputs: List[Dict[str, torch.Tensor]]
    ) -> Dict[str, Dict[str, torch.Tensor]]:
        """ Computes metrics. """
        return self.compute_metrics(outputs)

    def setup(self, stage) -> None:
        """Data preparation function called before training by Lightning.
        Equivalent to the prepare_data in previous Lightning Versions"""
        self.train_dataset = self.read_csv(self.hparams.train_path)
        self.val_dataset = self.read_csv(self.hparams.val_path)

        # Always validate the model with 2k examples from training to control overfit.
        train_subset = np.random.choice(a=len(self.train_dataset), size=2000)
        self.train_subset = Subset(self.train_dataset, train_subset)

        if self.hparams.test_path:
            self.test_dataset = self.read_csv(self.hparams.test_path)

    def train_dataloader(self) -> DataLoader:
        """ Function that loads the train set. """
        return DataLoader(
            dataset=self.train_dataset,
            sampler=RandomSampler(self.train_dataset),
            batch_size=self.hparams.batch_size,
            collate_fn=self.prepare_sample,
            num_workers=self.hparams.loader_workers,
        )

    def val_dataloader(self) -> DataLoader:
        """ Function that loads the validation set. """
        return [
            DataLoader(
                dataset=self.train_subset,
                batch_size=self.hparams.batch_size,
                collate_fn=self.prepare_sample,
                num_workers=self.hparams.loader_workers,
            ),
            DataLoader(
                dataset=self.val_dataset,
                batch_size=self.hparams.batch_size,
                collate_fn=self.prepare_sample,
                num_workers=self.hparams.loader_workers,
            ),
        ]

    def test_dataloader(self) -> DataLoader:
        """ Function that loads the validation set. """
        return DataLoader(
            dataset=self.test_dataset,
            batch_size=self.hparams.batch_size,
            collate_fn=self.prepare_sample,
            num_workers=self.hparams.loader_workers,
        )



# -*- coding: utf-8 -*-
r"""
Lightning Trainer Setup
==============
   Setup logic for the lightning trainer.
"""
import os
from argparse import Namespace
from datetime import datetime
from typing import Union

import click
import pandas as pd

import pytorch_lightning as pl
from comet.models.utils import apply_to_sample
from pytorch_lightning.callbacks import (
    Callback,
    EarlyStopping,
    ModelCheckpoint,
)
from pytorch_lightning.loggers import LightningLoggerBase, TensorBoardLogger
from pytorch_lightning.utilities import rank_zero_only


class TrainerConfig:
    """
    The TrainerConfig class is used to define default hyper-parameters that
    are used to initialize our Lightning Trainer. These parameters are then overwritted
    with the values defined in the YAML file.

    -------------------- General Parameters -------------------------

    :param seed: Training seed.

    :param deterministic: If true enables cudnn.deterministic. Might make your system
        slower, but ensures reproducibility.

    :param model: Model class we want to train.

    :param verbode: verbosity mode.

    :param overfit_batches: Uses this much data of the training set. If nonzero, will use
        the same training set for validation and testing. If the training dataloaders
        have shuffle=True, Lightning will automatically disable it.

    :param lr_finder: Runs a small portion of the training where the learning rate is increased
        after each processed batch and the corresponding loss is logged. The result of this is
        a lr vs. loss plot that can be used as guidance for choosing a optimal initial lr.

    -------------------- Model Checkpoint & Early Stopping -------------------------

    :param early_stopping: If true enables EarlyStopping.

    :param save_top_k: If save_top_k == k, the best k models according to the metric
        monitored will be saved.

    :param monitor: Metric to be monitored.

    :param save_weights_only: Saves only the weights of the model.

    :param period: Interval (number of epochs) between checkpoints.

    :param metric_mode: One of {min, max}. In min mode, training will stop when the
        metric monitored has stopped decreasing; in max mode it will stop when the
        metric monitored has stopped increasing.

    :param min_delta: Minimum change in the monitored metric to qualify as an improvement.

    :param patience: Number of epochs with no improvement after which training will be stopped.
    """

    seed: int = 3
    deterministic: bool = True
    model: str = None
    verbose: bool = False
    overfit_batches: Union[int, float] = 0.0

    # Model Checkpoint & Early Stopping
    early_stopping: bool = True
    save_top_k: int = 1
    monitor: str = "kendall"
    save_weights_only: bool = False
    metric_mode: str = "max"
    min_delta: float = 0.0
    patience: int = 1
    accumulate_grad_batches: int = 1
    lr_finder: bool = False

    def __init__(self, initial_data: dict) -> None:
        trainer_attr = pl.Trainer.default_attributes()
        for key in trainer_attr:
            setattr(self, key, trainer_attr[key])

        for key in initial_data:
            if hasattr(self, key):
                setattr(self, key, initial_data[key])

    def namespace(self) -> Namespace:
        return Namespace(
            **{
                name: getattr(self, name)
                for name in dir(self)
                if not callable(getattr(self, name)) and not name.startswith("__")
            }
        )


class TrainReport(Callback):
    """ Logger Callback that echos results during training. """

    _stack: list = []  # stack to keep metrics from all epochs

    @rank_zero_only
    def on_validation_end(
        self, trainer: pl.Trainer, pl_module: pl.LightningModule
    ) -> None:
        metrics = trainer.callback_metrics
        metrics = LightningLoggerBase._flatten_dict(metrics, "_")
        metrics = apply_to_sample(lambda x: x.item(), metrics)
        self._stack.append(metrics)
        # pl_module.print() # Print newline

    @rank_zero_only
    def on_fit_end(self, trainer: pl.Trainer) -> None:
        click.secho("\nTraining Report Experiment:", fg="yellow")
        index_column = ["Epoch " + str(i + 1) for i in range(len(self._stack))]
        df = pd.DataFrame(self._stack, index=index_column)
        click.secho("{}".format(df), fg="yellow")


def build_trainer(hparams: Namespace) -> pl.Trainer:
    """
    :param hparams: Namespace

    :returns: Lightning Trainer (obj)
    """
    # Early Stopping Callback
    early_stop_callback = EarlyStopping(
        monitor=hparams.monitor,
        min_delta=hparams.min_delta,
        patience=hparams.patience,
        verbose=hparams.verbose,
        mode=hparams.metric_mode,
    )

    # TestTube Logger Callback
    tb_logger = TensorBoardLogger(
        save_dir="experiments/",
        version="version_" + datetime.now().strftime("%d-%m-%Y--%H-%M-%S"),
        name="lightning",
    )

    # Model Checkpoint Callback
    ckpt_path = os.path.join(
        "experiments/lightning/",
        tb_logger.version,
        "checkpoints",
    )

    checkpoint_callback = ModelCheckpoint(
        filepath=ckpt_path,
        save_top_k=hparams.save_top_k,
        verbose=hparams.verbose,
        monitor=hparams.monitor,
        save_weights_only=hparams.save_weights_only,
        period=0,  # Always allow saving checkpoint even within the same epoch
        mode=hparams.metric_mode,
    )
    other_callbacks = [early_stop_callback, TrainReport()]

    trainer = pl.Trainer(
        logger=tb_logger,
        checkpoint_callback=checkpoint_callback,
        callbacks=other_callbacks,
        gradient_clip_val=hparams.gradient_clip_val,
        gpus=hparams.gpus,
        log_gpu_memory="all",
        deterministic=hparams.deterministic,
        overfit_batches=hparams.overfit_batches,
        check_val_every_n_epoch=1,
        fast_dev_run=False,
        accumulate_grad_batches=hparams.accumulate_grad_batches,
        max_epochs=hparams.max_epochs,
        min_epochs=hparams.min_epochs,
        limit_train_batches=hparams.limit_train_batches,
        limit_val_batches=hparams.limit_val_batches,
        val_check_interval=hparams.val_check_interval,
        distributed_backend=hparams.distributed_backend,
        precision=hparams.precision,
        weights_summary="top",
        profiler=hparams.profiler,
    )
    return trainer



# -*- coding: utf-8 -*-
r"""
Comet Ranker Model
======================
    The goal of this model is to rank good translations closer to the reference and source text
    and bad translations further by a small margin.

    https://pytorch.org/docs/stable/nn.html#tripletmarginloss
"""
from argparse import Namespace
from typing import Dict, List, Tuple, Union

import torch
import torch.nn.functional as F
from tqdm import tqdm

from comet.models.ranking.ranking_base import RankingBase
from comet.models.utils import move_to_cuda
from torchnlp.utils import collate_tensors


class CometRanker(RankingBase):
    """
    Comet Ranker class that uses a pretrained encoder to extract features
    from the sequences and then passes those features through a Triplet Margin Loss.

    :param hparams: Namespace containing the hyperparameters.
    """

    def __init__(self, hparams: Namespace) -> None:
        super().__init__(hparams)

    def compute_metrics(self, outputs: List[Dict[str, torch.Tensor]]) -> dict:
        """  Computes WMT19 shared task kendall tau like metric. """
        distance_pos, distance_neg = [], []
        for minibatch in outputs:
            minibatch = minibatch["val_prediction"]
            src_embedding = minibatch["src_sentemb"]
            ref_embedding = minibatch["ref_sentemb"]
            pos_embedding = minibatch["pos_sentemb"]
            neg_embedding = minibatch["neg_sentemb"]

            distance_src_pos = F.pairwise_distance(pos_embedding, src_embedding)
            distance_ref_pos = F.pairwise_distance(pos_embedding, ref_embedding)
            harmonic_distance_pos = (2 * distance_src_pos * distance_ref_pos) / (
                distance_src_pos + distance_ref_pos
            )
            distance_pos.append(harmonic_distance_pos)

            distance_src_neg = F.pairwise_distance(neg_embedding, src_embedding)
            distance_ref_neg = F.pairwise_distance(neg_embedding, ref_embedding)
            harmonic_distance_neg = (2 * distance_src_neg * distance_ref_neg) / (
                distance_src_neg + distance_ref_neg
            )
            distance_neg.append(harmonic_distance_neg)

        return {
            "kendall": self.metrics.compute(torch.cat(distance_pos), torch.cat(distance_neg))
        }

    def compute_loss(self, model_out: Dict[str, torch.Tensor], *args) -> torch.Tensor:
        """
        Computes Triplet Margin Loss for both the reference and the source.
        
        :param model_out: model specific output with src_anchor, ref_anchor, pos and neg
            sentence embeddings.
        """
        ref_anchor = model_out["ref_sentemb"]
        src_anchor = model_out["src_sentemb"]
        positive = model_out["pos_sentemb"]
        negative = model_out["neg_sentemb"]
        return self.loss(src_anchor, positive, negative) + self.loss(
            ref_anchor, positive, negative
        )

    def predict(
        self, samples: Dict[str, str], cuda: bool = False, show_progress: bool = False
    ) -> (Dict[str, Union[str, float]], List[float]):
        """ Function that runs a model prediction,
        
        :param samples: List of dictionaries with 'mt' and 'ref' keys.
        :param cuda: Flag that runs inference using 1 single GPU.
        :param show_progress: Flag to show progress during inference of multiple examples.
        
        :return: Dictionary with model outputs
        """
        if self.training:
            self.eval()

        if cuda and torch.cuda.is_available():
            self.to("cuda")

        with torch.no_grad():
            batches = [
                samples[i : i + self.hparams.batch_size]
                for i in range(0, len(samples), self.hparams.batch_size)
            ]
            model_inputs = []
            if show_progress:
                pbar = tqdm(
                    total=len(batches), desc="Preparing batches....", dynamic_ncols=True
                )
            for batch in batches:
                model_inputs.append(self.prepare_sample(batch, inference=True))
                if show_progress:
                    pbar.update(1)

            if show_progress:
                pbar.close()

            if show_progress:
                pbar = tqdm(
                    total=len(batches), desc="Scoring hypothesis...", dynamic_ncols=True
                )

            distance_weighted, distance_src, distance_ref = [], [], []
            for k, model_input in enumerate(model_inputs):
                src_input, mt_input, ref_input, alt_input = model_input
                if cuda and torch.cuda.is_available():
                    src_embeddings = self.get_sentence_embedding(
                        **move_to_cuda(src_input)
                    )
                    mt_embeddings = self.get_sentence_embedding(
                        **move_to_cuda(mt_input)
                    )
                    ref_embeddings = self.get_sentence_embedding(
                        **move_to_cuda(ref_input)
                    )
                    ref_distances = F.pairwise_distance(
                        mt_embeddings, ref_embeddings
                    ).cpu()
                    src_distances = F.pairwise_distance(
                        mt_embeddings, src_embeddings
                    ).cpu()

                    # When 2 references are given the distance to the reference is the Min between
                    # both references.
                    if alt_input is not None:
                        alt_embeddings = self.get_sentence_embedding(
                            **move_to_cuda(alt_input)
                        )
                        alt_distances = F.pairwise_distance(
                            mt_embeddings, alt_embeddings
                        ).cpu()
                        ref_distances = torch.stack([ref_distances, alt_distances])
                        ref_distances = ref_distances.min(dim=0).values

                else:
                    src_embeddings = self.get_sentence_embedding(**src_input)
                    mt_embeddings = self.get_sentence_embedding(**mt_input)
                    ref_embeddings = self.get_sentence_embedding(**ref_input)
                    ref_distances = F.pairwise_distance(mt_embeddings, ref_embeddings)
                    src_distances = F.pairwise_distance(mt_embeddings, src_embeddings)

                # Harmonic mean between the distances:
                distances = (2 * ref_distances * src_distances) / (
                    ref_distances + src_distances
                )
                src_distances = ref_distances.numpy().tolist()
                ref_distances = ref_distances.numpy().tolist()
                distances = distances.numpy().tolist()

                for i in range(len(distances)):
                    distance_weighted.append(1 / (1 + distances[i]))
                    distance_src.append(1 / (1 + src_distances[i]))
                    distance_ref.append(1 / (1 + ref_distances[i]))

                if show_progress:
                    pbar.update(1)

            if show_progress:
                pbar.close()

        assert len(distance_weighted) == len(samples)
        scores = [] 
        for i in range(len(samples)):
            scores.append(distance_weighted[i])
            samples[i]["predicted_score"] = scores[-1]
            samples[i]["reference_distance"] = distance_ref[i]
            samples[i]["source_distance"] = distance_src[i]

        return samples, scores

    def prepare_sample(
        self, sample: List[Dict[str, Union[str, float]]], inference: bool = False
    ) -> Union[Tuple[Dict[str, torch.Tensor], None], List[Dict[str, torch.Tensor]]]:
        """
        Function that prepares a sample to input the model.
        
        :param sample: list of dictionaries.
        :param inference: If set to to False, then the model expects 
            a MT and reference instead of anchor, pos, and neg segments.

        :return: Tuple with a dictionary containing the model inputs and None OR List 
            with source, MT and reference tokenized and vectorized.
        """
        sample = collate_tensors(sample)
        if inference:
            src_inputs = self.encoder.prepare_sample(sample["src"])
            mt_inputs = self.encoder.prepare_sample(sample["mt"])
            ref_inputs = self.encoder.prepare_sample(sample["ref"])
            alt_inputs = (
                self.encoder.prepare_sample(sample["alt"]) if "alt" in sample else None
            )
            return src_inputs, mt_inputs, ref_inputs, alt_inputs

        ref_inputs = self.encoder.prepare_sample(sample["ref"])
        src_inputs = self.encoder.prepare_sample(sample["src"])
        pos_inputs = self.encoder.prepare_sample(sample["pos"])
        neg_inputs = self.encoder.prepare_sample(sample["neg"])

        ref_inputs = {"ref_" + k: v for k, v in ref_inputs.items()}
        src_inputs = {"src_" + k: v for k, v in src_inputs.items()}
        pos_inputs = {"pos_" + k: v for k, v in pos_inputs.items()}
        neg_inputs = {"neg_" + k: v for k, v in neg_inputs.items()}

        return {**ref_inputs, **src_inputs, **pos_inputs, **neg_inputs}, torch.empty(0)

    def forward(
        self,
        src_tokens: torch.tensor,
        ref_tokens: torch.tensor,
        pos_tokens: torch.tensor,
        neg_tokens: torch.tensor,
        src_lengths: torch.tensor,
        ref_lengths: torch.tensor,
        pos_lengths: torch.tensor,
        neg_lengths: torch.tensor,
        **kwargs
    ) -> Dict[str, torch.Tensor]:
        """
        Function that encodes the anchor, positive samples and negative samples
        and returns embeddings for the triplet.

        :param src_tokens: anchor sequences [batch_size x anchor_seq_len]
        :param ref_tokens: anchor sequences [batch_size x anchor_seq_len]
        :param pos_tokens: positive sequences [batch_size x pos_seq_len]
        :param neg_tokens: negative sequences [batch_size x neg_seq_len]
        :param src_lengths: anchor lengths [batch_size]
        :param ref_lengths: anchor lengths [batch_size]
        :param pos_lengths: positive lengths [batch_size]
        :param neg_lengths: negative lengths [batch_size]

        :return: Dictionary with model outputs to be passed to the loss function.
        """
        return {
            "src_sentemb": self.get_sentence_embedding(src_tokens, src_lengths),
            "ref_sentemb": self.get_sentence_embedding(ref_tokens, ref_lengths),
            "pos_sentemb": self.get_sentence_embedding(pos_tokens, pos_lengths),
            "neg_sentemb": self.get_sentence_embedding(neg_tokens, neg_lengths),
        }



# -*- coding: utf-8 -*-
r"""
Comet Estimator Model
================================
    Comet Estimator predicts a quality score for the 
    hyphotesis (e.g: the MT text) by looking at reference, source and MT.
"""
import random
from argparse import Namespace
from typing import Dict, List, Tuple, Union

import torch

from comet.models.estimators.estimator_base import Estimator
from comet.modules.feedforward import FeedForward
from comet.modules.scalar_mix import ScalarMixWithDropout
from torchnlp.utils import collate_tensors


class CometEstimator(Estimator):
    """
    Estimator class that uses a pretrained encoder to extract features from
    the sequences and then passes those features to a feed forward estimator.

    :param hparams: Namespace containing the hyperparameters.
    """

    class ModelConfig(Estimator.ModelConfig):
        switch_prob: float = 0.0

    def __init__(
        self,
        hparams: Namespace,
    ) -> None:
        super().__init__(hparams)

    def _build_model(self) -> Estimator:
        """
        Initializes the estimator architecture.
        """
        super()._build_model()
        if self.hparams.encoder_model != "LASER":
            self.layer = (
                int(self.hparams.layer)
                if self.hparams.layer != "mix"
                else self.hparams.layer
            )

            self.scalar_mix = (
                ScalarMixWithDropout(
                    mixture_size=self.encoder.num_layers,
                    dropout=self.hparams.scalar_mix_dropout,
                    do_layer_norm=True,
                )
                if self.layer == "mix" and self.hparams.pool != "default"
                else None
            )

        input_emb_sz = (
            self.encoder.output_units * 6
            if self.hparams.pool != "cls+avg"
            else self.encoder.output_units * 2 * 6
        )

        self.ff = FeedForward(
            in_dim=input_emb_sz,
            hidden_sizes=self.hparams.hidden_sizes,
            activations=self.hparams.activations,
            dropout=self.hparams.dropout,
            final_activation=(
                self.hparams.final_activation
                if hasattr(
                    self.hparams, "final_activation"
                )  # compatability with older checkpoints!
                else "Sigmoid"
            ),
        )

    def configure_optimizers(
        self,
    ) -> Tuple[List[torch.optim.Optimizer], List[torch.optim.lr_scheduler.LambdaLR]]:
        """ Sets different Learning rates for different parameter groups. """
        layer_parameters = self.encoder.layerwise_lr(
            self.hparams.encoder_learning_rate, self.hparams.layerwise_decay
        )
        ff_parameters = [
            {"params": self.ff.parameters(), "lr": self.hparams.learning_rate}
        ]

        if self.hparams.encoder_model != "LASER" and self.scalar_mix:
            scalar_mix_parameters = [
                {
                    "params": self.scalar_mix.parameters(),
                    "lr": self.hparams.learning_rate,
                }
            ]

            optimizer = self._build_optimizer(
                layer_parameters + ff_parameters + scalar_mix_parameters
            )
        else:
            optimizer = self._build_optimizer(layer_parameters + ff_parameters)
        scheduler = self._build_scheduler(optimizer)
        return [optimizer], [scheduler]

    def prepare_sample(
        self, sample: List[Dict[str, Union[str, float]]], inference: bool = False
    ) -> Union[
        Tuple[Dict[str, torch.Tensor], Dict[str, torch.Tensor]], Dict[str, torch.Tensor]
    ]:
        """
        Function that prepares a sample to input the model.

        :param sample: list of dictionaries.
        :param inference: If set to true prepares only the model inputs.

        :returns: Tuple with 2 dictionaries (model inputs and targets). 
            If `inference=True` returns only the model inputs.
        """
        sample = collate_tensors(sample)
        src_inputs = self.encoder.prepare_sample(sample["src"])
        mt_inputs = self.encoder.prepare_sample(sample["mt"])
        ref_inputs = self.encoder.prepare_sample(sample["ref"])

        src_inputs = {"src_" + k: v for k, v in src_inputs.items()}
        mt_inputs = {"mt_" + k: v for k, v in mt_inputs.items()}
        ref_inputs = {"ref_" + k: v for k, v in ref_inputs.items()}

        if "alt" in sample:
            alt_inputs = self.encoder.prepare_sample(sample["alt"])
            alt_inputs = {"alt_" + k: v for k, v in alt_inputs.items()}
            inputs = {**src_inputs, **mt_inputs, **ref_inputs, **alt_inputs}

        else:
            inputs = {**src_inputs, **mt_inputs, **ref_inputs}

        if inference:
            return inputs

        targets = {"score": torch.tensor(sample["score"], dtype=torch.float)}
        return inputs, targets

    def forward(
        self,
        src_tokens: torch.tensor,
        mt_tokens: torch.tensor,
        ref_tokens: torch.tensor,
        src_lengths: torch.tensor,
        mt_lengths: torch.tensor,
        ref_lengths: torch.tensor,
        alt_tokens: torch.tensor = None,
        alt_lengths: torch.tensor = None,
        **kwargs
    ) -> Dict[str, torch.Tensor]:
        """
        Function that encodes both Source, MT and Reference and returns a quality score.

        :param src_tokens: SRC sequences [batch_size x src_seq_len]
        :param mt_tokens: MT sequences [batch_size x mt_seq_len]
        :param ref_tokens: REF sequences [batch_size x ref_seq_len]
        :param src_lengths: SRC lengths [batch_size]
        :param mt_lengths: MT lengths [batch_size]
        :param ref_lengths: REF lengths [batch_size]

        :param alt_tokens: Alternative REF sequences [batch_size x alt_seq_len]
        :param alt_lengths: Alternative REF lengths [batch_size]

        :return: Dictionary with model outputs to be passed to the loss function.
        """
        src_sentemb = self.get_sentence_embedding(src_tokens, src_lengths)
        mt_sentemb = self.get_sentence_embedding(mt_tokens, mt_lengths)
        ref_sentemb = self.get_sentence_embedding(ref_tokens, ref_lengths)

        diff_ref = torch.abs(mt_sentemb - ref_sentemb)
        diff_src = torch.abs(mt_sentemb - src_sentemb)

        prod_ref = mt_sentemb * ref_sentemb
        prod_src = mt_sentemb * src_sentemb

        if (
            not hasattr(
                self.hparams, "switch_prob"
            )  # compatability with older checkpoints!
            or self.hparams.switch_prob <= 0.0
        ):
            embedded_sequences = torch.cat(
                (mt_sentemb, ref_sentemb, prod_ref, diff_ref, prod_src, diff_src), dim=1
            )
            score = self.ff(embedded_sequences)

            if (alt_tokens is not None) and (alt_lengths is not None):

                alt_sentemb = self.get_sentence_embedding(alt_tokens, alt_lengths)

                diff_alt = torch.abs(mt_sentemb - alt_sentemb)
                prod_alt = mt_sentemb * alt_sentemb

                embedded_sequences = torch.cat(
                    (mt_sentemb, alt_sentemb, prod_alt, diff_alt, prod_src, diff_src),
                    dim=1,
                )
                score = (score + self.ff(embedded_sequences)) / 2

            return {"score": score}

        if self.training:
            switch = random.random() < self.hparams.switch_prob

            if switch:
                embedded_sequences = torch.cat(
                    (mt_sentemb, ref_sentemb, prod_src, diff_src, prod_ref, diff_ref),
                    dim=1,
                )
            else:
                embedded_sequences = torch.cat(
                    (mt_sentemb, ref_sentemb, prod_ref, diff_ref, prod_src, diff_src),
                    dim=1,
                )
            return {"score": self.ff(embedded_sequences)}

        elif (alt_tokens is not None) and (alt_lengths is not None):
            # Switcheroo Inference!
            alt_sentemb = self.get_sentence_embedding(alt_tokens, alt_lengths)
            diff_alt = torch.abs(mt_sentemb - alt_sentemb)
            prod_alt = mt_sentemb * alt_sentemb

            # Source + MT + Reference
            embedded_sequences = torch.cat(
                (mt_sentemb, ref_sentemb, prod_ref, diff_ref, prod_src, diff_src), dim=1
            )
            src_mt_ref = self.ff(embedded_sequences)

            # Reference + MT + Source
            embedded_sequences = torch.cat(
                (mt_sentemb, ref_sentemb, prod_src, diff_src, prod_ref, diff_ref), dim=1
            )
            ref_mt_src = self.ff(embedded_sequences)

            # Source + MT + Alternative Reference
            embedded_sequences = torch.cat(
                (mt_sentemb, alt_sentemb, prod_alt, diff_alt, prod_src, diff_src), dim=1
            )
            src_mt_alt = self.ff(embedded_sequences)

            # Alternative Reference + MT + Source
            embedded_sequences = torch.cat(
                (mt_sentemb, alt_sentemb, prod_src, diff_src, prod_alt, diff_alt), dim=1
            )
            alt_mt_src = self.ff(embedded_sequences)

            # Alternative Reference + MT + Reference
            embedded_sequences = torch.cat(
                (mt_sentemb, alt_sentemb, prod_alt, diff_alt, prod_ref, diff_ref), dim=1
            )
            alt_mt_ref = self.ff(embedded_sequences)

            # Reference + MT + Alternative Reference
            embedded_sequences = torch.cat(
                (mt_sentemb, ref_sentemb, prod_ref, diff_ref, prod_alt, diff_alt), dim=1
            )
            ref_mt_alt = self.ff(embedded_sequences)

            score = torch.stack(
                [src_mt_ref, ref_mt_src, src_mt_alt, alt_mt_src, alt_mt_ref, ref_mt_alt]
            )
            confidence = 1 - score.std(dim=0)

            return {"score": score.mean(dim=0) * confidence, "confidence": confidence}

        else:
            # Usual scoring
            embedded_sequences = torch.cat(
                (mt_sentemb, ref_sentemb, prod_ref, diff_ref, prod_src, diff_src), dim=1
            )
            score = self.ff(embedded_sequences) * (1 - self.hparams.switch_prob)

            # Switch src and reference embeddings
            embedded_sequences = torch.cat(
                (mt_sentemb, ref_sentemb, prod_src, diff_src, prod_ref, diff_ref), dim=1
            )
            return {
                "score": score + self.ff(embedded_sequences) * self.hparams.switch_prob
            }
